'''
Created on Dec 31, 2016

@author: nicolas
'''

import os

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import get_suites_from_project
from lemoncheesecake.exceptions import LemonCheesecakeException, ProgrammingError, UserError, \
    serialize_current_exception
from lemoncheesecake.filter import add_run_filter_cli_args
from lemoncheesecake.fixtures import FixtureRegistry, BuiltinFixture
from lemoncheesecake.project import find_project_file, load_project_from_file, load_project
from lemoncheesecake.reporting import filter_reporting_backends_by_capabilities, CAPABILITY_REPORTING_SESSION
from lemoncheesecake.runner import run_suites
from lemoncheesecake import events


def build_fixture_registry(project, cli_args):
    registry = FixtureRegistry()
    registry.add_fixture(BuiltinFixture("cli_args", lambda: cli_args))
    registry.add_fixture(BuiltinFixture("project_dir", lambda: project.get_project_dir()))
    for fixture in project.get_fixtures():
        registry.add_fixture(fixture)
    registry.check_dependencies()
    return registry


class RunCommand(Command):
    def get_name(self):
        return "run"

    def get_description(self):
        return "Run the tests"

    def add_cli_args(self, cli_parser):
        project_file = find_project_file()
        project = None
        default_reporting_backend_names = []
        if project_file:
            project = load_project_from_file(project_file)
            default_reporting_backend_names = [backend.name for backend in project.get_default_reporting_backends_for_test_run()]

        add_run_filter_cli_args(cli_parser)

        test_execution_group = cli_parser.add_argument_group("Test execution")
        test_execution_group.add_argument("--exit-error-on-failure", action="store_true",
            help="Exit with non-zero code if there is at least one non-passed test"
        )
        test_execution_group.add_argument("--stop-on-failure", action="store_true",
            help="Stop tests execution on the first non-passed test"
        )
        reporting_group = cli_parser.add_argument_group("Reporting")
        reporting_group.add_argument("--report-dir", "-r", required=False, help="Directory where report data will be stored")
        reporting_group.add_argument("--reporting", nargs="+", default=default_reporting_backend_names,
            help="The list of reporting backends to use"
        )
        reporting_group.add_argument("--enable-reporting", nargs="+", default=[],
            help="The list of reporting backends to add (to base backends)"
        )
        reporting_group.add_argument("--disable-reporting", nargs="+", default=[],
            help="The list of reporting backends to remove (from base backends)"
        )

        if project:
            project.add_custom_args_to_run_cli(cli_parser)

    def run_cmd(self, cli_args):
        # Project initialization
        project = load_project()
        suites = get_suites_from_project(project, cli_args)
        events.add_listener(project)

        # Build fixture registry
        fixture_registry = build_fixture_registry(project, cli_args)
        fixture_registry.check_fixtures_in_suites(suites)

        # Set reporting backends
        reporting_backends = {
            backend.name: backend for backend in
                filter_reporting_backends_by_capabilities(project.get_all_reporting_backends(), CAPABILITY_REPORTING_SESSION)
        }
        selected_reporting_backends = set()
        for backend_name in cli_args.reporting + cli_args.enable_reporting:
            try:
                selected_reporting_backends.add(reporting_backends[backend_name])
            except KeyError:
                raise LemonCheesecakeException("Unknown reporting backend '%s'" % backend_name)
        for backend_name in cli_args.disable_reporting:
            try:
                selected_reporting_backends.discard(reporting_backends[backend_name])
            except KeyError:
                raise LemonCheesecakeException("Unknown reporting backend '%s'" % backend_name)

        # Create report dir
        if cli_args.report_dir:
            report_dir = cli_args.report_dir
            try:
                os.mkdir(report_dir)
            except Exception as e:
                return LemonCheesecakeException("Cannot create report directory: %s" % e)
        else:
            try:
                report_dir = project.create_report_dir()
            except UserError as e:
                raise e
            except Exception:
                raise LemonCheesecakeException(
                    "Got an unexpected exception while creating report directory:%s" % \
                    serialize_current_exception(show_stacktrace=True)
                )

        # Handle before run hook
        try:
            project.run_pre_session_hook(report_dir)
        except UserError as e:
            raise e
        except Exception:
            raise ProgrammingError(
                "Got an unexpected exception while running the pre-session hook:%s" % \
                serialize_current_exception(show_stacktrace=True)
            )

        # Run tests
        is_successful = run_suites(
            suites, fixture_registry, selected_reporting_backends, report_dir,
            stop_on_failure=cli_args.stop_on_failure
        )

        # Handle after run hook
        try:
            project.run_post_session_hook(report_dir)
        except UserError as e:
            raise e
        except Exception:
            raise ProgrammingError(
                "Got an unexpected exception while running the post-session hook:%s" % \
                serialize_current_exception(show_stacktrace=True)
            )

        if cli_args.exit_error_on_failure:
            return 0 if is_successful else 1
        else:
            return 0



from functools import reduce

from lemoncheesecake.utils import humanize_duration
from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends, add_report_path_cli_arg, get_report_path
from lemoncheesecake.cli.display import print_table
from lemoncheesecake.reporting import load_report
from lemoncheesecake.filter import add_report_filter_cli_args, make_report_filter, filter_suites
from lemoncheesecake.testtree import flatten_tests, flatten_suites
from lemoncheesecake.reporting.report import flatten_steps


def get_tests_ordered_by_duration(tests):
    return sorted(tests, key=lambda test: test.duration, reverse=True)


def get_suites_ordered_by_duration(suites):
    return sorted(
        filter(lambda suite: len(suite.get_tests()) > 0, suites),
        key=lambda suite: suite.duration,
        reverse=True
    )


def get_total_duration(elems):
    return reduce(lambda x, y: x + y, [elem.duration for elem in elems], 0)


def format_test_entry(elem, total_time):
    if elem.duration is not None:
        return (
            elem.path,
            humanize_duration(elem.duration, show_milliseconds=True),
            "%d%%" % (elem.duration / total_time * 100)
        )
    else:
        return elem.path, "-", "-"


def get_top_tests(suites):
    tests = get_tests_ordered_by_duration(flatten_tests(suites))
    total_duration = get_total_duration(tests)
    return [format_test_entry(test, total_duration) for test in tests]


class TopTests(Command):
    def get_name(self):
        return "top-tests"

    def get_description(self):
        return "Display tests ordered by duration"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Top tests")
        add_report_path_cli_arg(group)
        add_report_filter_cli_args(cli_parser, no_positional_argument=True)

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)

        report = load_report(report_path, auto_detect_reporting_backends())
        suites = filter_suites(report.get_suites(), make_report_filter(cli_args))

        print_table(
            "Tests, ordered by duration",
            ("Test", "Duration", "In %"),
            get_top_tests(suites)
        )

        return 0


def get_top_suites(suites):
    processed_suites = get_suites_ordered_by_duration(flatten_suites(suites))
    total_duration = get_total_duration(processed_suites)
    return [format_test_entry(suite, total_duration) for suite in processed_suites]


class TopSuites(Command):
    def get_name(self):
        return "top-suites"

    def get_description(self):
        return "Display suites ordered by duration"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Top suites")
        add_report_path_cli_arg(group)
        add_report_filter_cli_args(cli_parser, no_positional_argument=True)

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)

        report = load_report(report_path, auto_detect_reporting_backends())
        suites = filter_suites(report.get_suites(), make_report_filter(cli_args))

        print_table(
            "Suites, ordered by duration",
            ("Suite", "Duration", "In %"),
            get_top_suites(suites)
        )

        return 0


def group_steps_by_description(steps):
    steps_by_description = {}
    for step in steps:
        if step.description not in steps_by_description:
            steps_by_description[step.description] = []
        steps_by_description[step.description].append(step)
    return steps_by_description


def get_steps_min_duration(steps):
    return min([step.duration for step in steps])


def get_steps_max_duration(steps):
    return max([step.duration for step in steps])


def get_steps_average_duration(steps):
    return get_total_duration(steps) / len(steps)


def format_steps_aggregation(description, occurrences, min_duration, max_duration, average_duration, duration, duration_pct):
    return (
        description,
        str(occurrences),
        humanize_duration(min_duration, show_milliseconds=True),
        humanize_duration(max_duration, show_milliseconds=True),
        humanize_duration(average_duration, show_milliseconds=True),
        humanize_duration(duration, show_milliseconds=True),
        "%d%%" % duration_pct
    )


def get_steps_aggregation(steps):
    steps_by_description = group_steps_by_description(steps)
    total_duration = get_total_duration(steps)

    data = []
    for description, steps in steps_by_description.items():
        duration = get_total_duration(steps)
        data.append([
            description,
            len(steps),
            get_steps_min_duration(steps),
            get_steps_max_duration(steps),
            get_steps_average_duration(steps),
            duration,
            duration / total_duration * 100
        ])

    return sorted(data, key=lambda row: row[-2], reverse=True)


def get_top_steps(suites):
    steps_aggregation = get_steps_aggregation(list(flatten_steps(suites)))
    return [format_steps_aggregation(*agg) for agg in steps_aggregation]


class TopSteps(Command):
    def get_name(self):
        return "top-steps"

    def get_description(self):
        return "Display steps aggregated and ordered by duration"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Top steps")
        add_report_path_cli_arg(group)
        add_report_filter_cli_args(cli_parser, no_positional_argument=True)

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)

        report = load_report(report_path, auto_detect_reporting_backends())
        suites = filter_suites(report.get_suites(), make_report_filter(cli_args))

        print_table(
            "Steps, aggregated and ordered by duration",
            ("Step", "Occ.", "Min.", "Max", "Avg.", "Total", "In %"),
            get_top_steps(suites)
        )

        return 0



'''
Created on Feb 5, 2017

@author: nicolas
'''

import os.path as osp
import inspect

from lemoncheesecake.importer import get_matching_files, get_py_files_from_dir, strip_py_ext, import_module
from lemoncheesecake.exceptions import UserError, ProgrammingError, ModuleImportError, InvalidMetadataError, \
    InvalidSuiteError, VisibilityConditionNotMet, serialize_current_exception
from lemoncheesecake.suite.core import Test, Suite, SUITE_HOOKS

__all__ = "load_suite_from_file", "load_suites_from_files", "load_suites_from_directory", \
    "load_suite_from_class", "load_suites_from_classes"


def is_suite_class(obj):
    return inspect.isclass(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_suite


def is_test_method(obj):
    return inspect.ismethod(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test


def is_test_function(obj):
    return inspect.isfunction(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test


def ensure_node_is_visible(obj, metadata):
    if metadata.condition is not None and not metadata.condition(obj):
        raise VisibilityConditionNotMet()


def load_test(obj):
    md = obj._lccmetadata
    ensure_node_is_visible(obj, md)

    test = Test(md.name, md.description, obj)
    test.tags.extend(md.tags)
    test.properties.update(md.properties)
    test.links.extend(md.links)
    test.disabled = md.disabled
    return test


def load_test_from_method(method):
    return load_test(method)


def load_test_from_function(func):
    return load_test(func)


def load_tests(objs):
    for obj in objs:
        try:
            yield load_test(obj)
        except VisibilityConditionNotMet:
            pass


def load_tests_from_methods(methods):
    return load_tests(methods)


def load_tests_from_functions(funcs):
    return load_tests(funcs)


def _list_object_attributes(obj):
    return [getattr(obj, n) for n in dir(obj) if not n.startswith("__")]


def _get_test_symbols(obj, filter_func):
    return sorted(filter(filter_func, _list_object_attributes(obj)), key=lambda sym: sym._lccmetadata.rank)


def get_test_methods_from_class(obj):
    return _get_test_symbols(obj, is_test_method)


def get_sub_suites_from_class(obj):
    return _get_test_symbols(obj, is_suite_class)


def get_test_functions_from_module(mod):
    return _get_test_symbols(mod, is_test_function)


def get_suite_classes_from_module(mod):
    return _get_test_symbols(mod, is_suite_class)


def get_generated_tests(obj):
    return getattr(obj, "_lccgeneratedtests", [])


def load_suite_from_class(class_):
    try:
        md = class_._lccmetadata
    except AttributeError:
        raise InvalidSuiteError()

    try:
        suite_obj = class_()
    except UserError as e:
        raise e # propagate UserError
    except Exception:
        raise ProgrammingError("Got an unexpected error while instantiating suite class '%s':%s" % (
            class_.__name__, serialize_current_exception()
        ))
    ensure_node_is_visible(suite_obj, md)

    suite = Suite(suite_obj, md.name, md.description)
    suite.tags.extend(md.tags)
    suite.properties.update(md.properties)
    suite.links.extend(md.links)
    suite.rank = md.rank
    suite.disabled = md.disabled

    for hook_name in SUITE_HOOKS:
        if hasattr(suite_obj, hook_name):
            suite.add_hook(hook_name, getattr(suite_obj, hook_name))

    for test in load_tests_from_methods(get_test_methods_from_class(suite_obj)):
        suite.add_test(test)

    for test in get_generated_tests(suite_obj):
        suite.add_test(test)

    for sub_suite in load_suites_from_classes(get_sub_suites_from_class(suite_obj)):
        suite.add_suite(sub_suite)

    return suite


def load_suites_from_classes(classes):
    suites = []
    for class_ in classes:
        try:
            suites.append(load_suite_from_class(class_))
        except VisibilityConditionNotMet:
            pass
    return suites


def load_suite_from_module(mod):
    # TODO: find a better way to workaround circular import
    from lemoncheesecake.suite.definition import get_metadata_next_rank

    suite_info = getattr(mod, "SUITE")
    suite_condition = suite_info.get("conditional")
    if suite_condition is not None and not suite_condition(mod):
        raise VisibilityConditionNotMet()

    suite_name = inspect.getmodulename(inspect.getfile(mod))

    try:
        suite_description = suite_info["description"]
    except KeyError:
        raise InvalidMetadataError("Missing description in '%s' suite information" % mod.__file__)

    suite = Suite(mod, suite_name, suite_description)
    suite.tags.extend(suite_info.get("tags", []))
    suite.properties.update(suite_info.get("properties", []))
    suite.links.extend(suite_info.get("links", []))
    suite.rank = suite_info.get("rank", get_metadata_next_rank())

    for hook_name in SUITE_HOOKS:
        if hasattr(mod, hook_name):
            suite.add_hook(hook_name, getattr(mod, hook_name))

    for test in load_tests_from_functions(get_test_functions_from_module(mod)):
        suite.add_test(test)

    for test in get_generated_tests(suite):
        suite.add_test(test)

    for sub_suite in load_suites_from_classes(get_suite_classes_from_module(mod)):
        suite.add_suite(sub_suite)

    return suite


def load_suite_from_file(filename):
    """Get suite from Python module.

    A valid module is either:
    - a module containing a dict name 'SUITE' with keys:
      - description (mandatory)
      - tags (optional)
      - properties (optional)
      - links (optional)
      - rank (optional)
    - a module that contains a suite class with the same name as the module name

    Raise a ModuleImportError if the suite class cannot be imported.
    """
    mod = import_module(filename)

    if hasattr(mod, "SUITE"):
        suite = load_suite_from_module(mod)
    else:
        mod_name = strip_py_ext(osp.basename(filename))
        try:
            klass = getattr(mod, mod_name)
        except AttributeError:
            raise ModuleImportError("Cannot find class '%s' in '%s'" % (mod_name, mod.__file__))

        suite = load_suite_from_class(klass)
    return suite


def load_suites_from_files(patterns, excluding=[]):
    """
    Import suites from a list of files:
    - patterns: a mandatory list (a simple string can also be used instead of a single element list)
      of files to import; the wildcard '*' character can be used
    - exclude: an optional list (a simple string can also be used instead of a single element list)
      of elements to exclude from the expanded list of files to import
    Example: load_suites_from_files("test_*.py")
    """
    suites = []
    for filename in get_matching_files(patterns, excluding):
        try:
            suites.append(load_suite_from_file(filename))
        except VisibilityConditionNotMet:
            pass
    return suites


def load_suites_from_directory(dir, recursive=True):
    """Find suite classes in modules found in dir.

    The function expect that:
    - each module (.py file) contains a class that inherits Suite
    - the class name must have the same name as the module name (if the module is foo.py
      the class must be named foo)
    If the recursive argument is set to True, sub suites will be searched in a directory named
    from the suite module: if the suite module is "foo.py" then the sub suites directory must be "foo".

    Raise ModuleImportError if one or more suite cannot be imported.
    """
    if not osp.exists(dir):
        raise ModuleImportError("Directory '%s' does not exist" % dir)

    suites = []
    for filename in get_py_files_from_dir(dir):
        try:
            suite = load_suite_from_file(filename)
        except VisibilityConditionNotMet:
            continue

        if recursive:
            subsuites_dir = strip_py_ext(filename)
            if osp.isdir(subsuites_dir):
                for sub_suite in load_suites_from_directory(subsuites_dir, recursive=True):
                    suite.add_suite(sub_suite)
        suites.append(suite)
    suites.sort(key=lambda suite: suite.rank)
    return suites



'''
Created on Apr 3, 2017

@author: nicolas
'''

from lemoncheesecake.utils import IS_PYTHON3, IS_PYTHON2
from lemoncheesecake.matching.base import Matcher, match_success, match_failure, got, serialize_value, got_value, to_be
from lemoncheesecake.matching.matchers.value import is_

__all__ = (
    "is_integer", "is_float", "is_bool", "is_str", "is_dict", "is_list"
)

TYPE_NAMES = {
    int: "integer",
    float: "float",
    str: "string",
    dict: "collection",
    list: "array", tuple: "array"
}

if IS_PYTHON2:
    TYPE_NAMES[unicode] = "string",


def get_value_type_name(value):
    try:
        return TYPE_NAMES[type(value)]
    except KeyError:
        return str(type(value))


class IsValueOfType(Matcher):
    def __init__(self, types, type_name, value_matcher):
        self.types = types
        self.type_name = type_name
        self.value_matcher = value_matcher

    def description(self, conjugate=False):
        ret = "%s %s" % (to_be(conjugate), self.type_name)
        if self.value_matcher:
            ret += " that %s" % self.value_matcher.description(conjugate=True)

        return ret

    def matches(self, actual):
        if type(actual) in self.types:
            if self.value_matcher:
                return self.value_matcher.matches(actual)
            else:
                return match_success(got_value(actual))
        else:
            return match_failure(got("%s (%s)" % (serialize_value(actual), get_value_type_name(actual))))


def is_type(types, type_name):
    def wrapper(value_matcher=None):
        return IsValueOfType(
            types, type_name, is_(value_matcher) if value_matcher is not None else None
        )
    wrapper.__doc__ = "Test if value is of type %s" % type_name
    return wrapper


is_integer = is_type([int], "an integer")
is_float = is_type([float], "a float")
is_bool = is_type([bool], "a boolean")
is_str = is_type([str] if IS_PYTHON3 else [str, unicode], "a string")
is_dict = is_type([dict], "a collection")
is_list = is_type([list, tuple], "a list")



from .console import ConsoleBackend
from .xml import XmlBackend
from .json_ import JsonBackend
from .html import HtmlBackend
from .junit import JunitBackend
from .reportportal import ReportPortalBackend
from .slack import SlackReportingBackend



'''
Created on Mar 27, 2017

@author: nicolas
'''

import json

from lemoncheesecake.utils import get_distincts_in_list
from lemoncheesecake.exceptions import method_not_implemented


class MatchResult:
    def __init__(self, outcome, description):
        self.outcome = outcome
        self.description = description

    def is_success(self):
        return self.outcome == True

    def is_failure(self):
        return self.outcome == False


def match_success(description=None):
    return MatchResult(True, description)


def match_failure(description):
    return MatchResult(False, description)


def match_result(outcome, description=None):
    return MatchResult(outcome, description)


class Matcher:
    def description(self, conjugate=False):
        method_not_implemented("description", self)

    def short_description(self, conjugate=False):
        return self.description(conjugate=conjugate)

    def matches(self, actual):
        method_not_implemented("match", self)


class MatchExpected(Matcher):
    def __init__(self, expected):
        self.expected = expected


def serialize_value(value):
    return json.dumps(value, ensure_ascii=False)


def serialize_values(values):
    return ", ".join(map(serialize_value, values))


def got(value=None):
    ret = "got"
    if value is not None:
        ret += " " + value
    return ret


def got_value(value):
    return got(serialize_value(value))


def got_values(values):
    return got(serialize_values(values))


def merge_match_result_descriptions(results):
    return ", ".join(
        get_distincts_in_list([result.description for result in results if result.description != None])
    )


def to_be(conjugate=False, negation=False):
    negative_form = " not" if negation else ""
    return "is%s" % negative_form if conjugate else "to%s be" % negative_form


def to_have(conjugate=False, negation=False):
    negative_form = " not" if negation else ""
    return "has%s" % negative_form if conjugate else "to%s have" % negative_form


def to_meet(conjugate=False, negation=False):
    if conjugate:
        if negation:
            return "does not meet"
        else:
            return "meets"
    else:
        if negation:
            return "to not meet"
        else:
            return "to meet"



'''
Created on Mar 27, 2017

@author: nicolas
'''

from lemoncheesecake.utils import is_string
from lemoncheesecake.runtime import log_check
from lemoncheesecake.exceptions import AbortTest, ProgrammingError
from lemoncheesecake.matching.base import Matcher
from lemoncheesecake.matching.matchers.dict_ import HasEntry, wrap_key_matcher
from lemoncheesecake.matching.matchers.composites import is_

__all__ = (
    "log_match_result",
    "check_that", "check_that_entry", "check_that_in",
    "require_that", "require_that_entry", "require_that_in",
    "assert_that", "assert_that_entry", "assert_that_in",
    "this_dict"
)


class _HasEntry(HasEntry):
    def description(self):
        ret = 'entry %s' % self.key_matcher.description()
        if self.value_matcher:
            ret += " " + self.value_matcher.description()
        return ret


class DictContext:
    def __init__(self, actual, base_key):
        self.actual = actual
        self.base_key = base_key


class ThisDict:
    _contexts = []

    def __init__(self, actual):
        self.context = DictContext(actual, [])

    def using_base_key(self, base_key):
        self.context.base_key = base_key if type(base_key) in (list, tuple) else [base_key]
        return self

    def __enter__(self):
        self._contexts.append(self.context)
        return self.context.actual

    def __exit__(self, type_, value, traceback):
        self._contexts.pop()

    @staticmethod
    def get_current_context():
        try:
            return ThisDict._contexts[-1]
        except IndexError:
            return None


def this_dict(actual):
    """
    Set actual to be used as implicit dict in *_that_entry functions when used
    within a 'with' statement.
    """
    return ThisDict(actual)


def _is_matcher(obj):
    return isinstance(obj, Matcher)


def _get_actual_for_dict_operation(in_):
    if in_ is not None:
        return in_
    else:
        context = ThisDict.get_current_context()
        if not context:
            raise ProgrammingError(
                "Actual dict must be set either using in_ argument or using 'with this_dict(...)' statement"
            )
        return context.actual


def _get_matcher_for_dict_operation(key_matcher, value_matcher, base_key):
    if base_key is None:
        context = ThisDict.get_current_context()
        base_key = context.base_key if context is not None else []

    return _HasEntry(
        wrap_key_matcher(key_matcher, base_key=base_key),
        value_matcher if value_matcher is not None else is_(value_matcher)
    )


def format_result_details(details):
    if details is None:
        return None

    if not is_string(details):
        details = str(details)

    return details[0].upper() + details[1:]


def log_match_result(hint, matcher, result, quiet=False):
    """Add a check log to the report.

    If quiet is set to True, the check details won't appear in the check log.
    """
    if hint is not None:
        description = "Expect %s %s" % (hint, matcher.description())
    else:
        description = "Expect %s" % matcher.description()

    return log_check(
        description, result.outcome, format_result_details(result.description) if not quiet else None
    )


def check_that(hint, actual, matcher, quiet=False):
    """Check that actual matches given matcher.

    A check log is added to the report.

    If quiet is set to True, the check details won't appear in the check log.
    """
    assert _is_matcher(matcher)

    result = matcher.matches(actual)
    log_match_result(hint, matcher, result, quiet=quiet)
    return result.outcome


def check_that_entry(key_matcher, value_matcher=None, in_=None, base_key=None, quiet=False):
    """
    Helper function for check_that, takes the actual dict using in_ parameter or using 'with this_dict(...)' statement
    """
    assert value_matcher is None or _is_matcher(value_matcher)

    return check_that(
        None,
        _get_actual_for_dict_operation(in_),
        _get_matcher_for_dict_operation(key_matcher, value_matcher, base_key),
        quiet=quiet
    )


def _do_that_in(func, actual, *args, **kwargs):
    if len(args) % 2 != 0:
        raise TypeError("function expects an even number of *args")

    base_key = kwargs.get("base_key", [])
    quiet = kwargs.get("quiet", False)

    i = 0
    while i < len(args):
        key, value_matcher = args[i], args[i+1]
        func(key, value_matcher, in_=actual, base_key=base_key, quiet=quiet)
        i += 2


def check_that_in(actual, *args, **kwargs):
    _do_that_in(check_that_entry, actual, *args, **kwargs)


def require_that_in(actual, *args, **kwargs):
    _do_that_in(require_that_entry, actual, *args, **kwargs)


def assert_that_in(actual, *args, **kwargs):
    _do_that_in(assert_that_entry, actual, *args, **kwargs)


def require_that(hint, actual, matcher, quiet=False):
    """Require that actual matches given matcher.

    A check log is added to the report. An AbortTest exception is raised if the check does not succeed.

    If quiet is set to True, the check details won't appear in the check log.
    """
    assert _is_matcher(matcher)

    result = matcher.matches(actual)
    log_match_result(hint, matcher, result, quiet=quiet)
    if result.is_failure():
        raise AbortTest("previous requirement was not fulfilled")


def require_that_entry(key_matcher, value_matcher=None, in_=None, base_key=None, quiet=False):
    """
    Helper function for require_that, takes the actual dict using in_ parameter or using 'with this_dict(...)' statement
    """
    assert value_matcher is None or _is_matcher(value_matcher)

    return require_that(
        None,
        _get_actual_for_dict_operation(in_),
        _get_matcher_for_dict_operation(key_matcher, value_matcher, base_key),
        quiet=quiet
    )


def assert_that(hint, actual, matcher, quiet=False):
    """Assert that actual matches given matcher.

    If assertion fail, a check log is added to the report and an AbortTest exception is raised.

    If quiet is set to True, the check details won't appear in the check log.
    """
    assert _is_matcher(matcher)

    result = matcher.matches(actual)
    if result.is_failure():
        log_match_result(hint, matcher, result, quiet=quiet)
        raise AbortTest("assertion error")


def assert_that_entry(key_matcher, value_matcher=None, in_=None, base_key=None, quiet=False):
    """
    Helper function for assert_that, takes the actual dict using in_ parameter or using 'with this_dict(...)' statement
    """
    assert value_matcher is None or _is_matcher(value_matcher)

    return assert_that(
        None,
        _get_actual_for_dict_operation(in_),
        _get_matcher_for_dict_operation(key_matcher, value_matcher, base_key),
        quiet=quiet
    )



'''
Created on Mar 27, 2016

@author: nicolas
'''

try:
    from lxml import etree as ET
    from lxml.builder import E
    LXML_IS_AVAILABLE = True
except ImportError:
    LXML_IS_AVAILABLE = False

import lemoncheesecake
from lemoncheesecake.reporting.backend import BoundReport, FileReportBackend, SAVE_AT_EACH_FAILED_TEST
from lemoncheesecake.reporting.report import (
    LogData, CheckData, AttachmentData, UrlData, StepData, TestData, HookData, SuiteData,
    format_timestamp, parse_timestamp
)
from lemoncheesecake.utils import IS_PYTHON3
from lemoncheesecake.exceptions import ProgrammingError, InvalidReportFile

OUTCOME_NOT_AVAILABLE = "n/a"
OUTCOME_FAILURE = "failure"
OUTCOME_SUCCESS = "success"

DEFAULT_INDENT_LEVEL = 4


# borrowed from http://stackoverflow.com/a/1239193
def indent_xml(elem, level=0, indent_level=DEFAULT_INDENT_LEVEL):
    i = "\n" + level * (" " * indent_level)
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + (" " * indent_level)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent_xml(elem, level+1, indent_level)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def set_node_attr(node, attr_name, attr_value):
    if IS_PYTHON3:
        node.attrib[attr_name] = attr_value
    else:
        node.attrib[attr_name] = attr_value if type(attr_value) is unicode else unicode(attr_value, "utf-8")


def make_xml_node(name, *args):
    node = E(name)
    i = 0
    while i < len(args):
        attr_name, attr_value = args[i], args[i+1]
        if attr_value is not None:
            set_node_attr(node, attr_name, attr_value)
        i += 2
    return node


def make_xml_child(parent_node, name, *args):
    node = make_xml_node(name, *args)
    parent_node.append(node)
    return node


def _add_time_attr(node, name, value):
    if not value:
        return
    node.attrib[name] = format_timestamp(value)


def _serialize_outcome(outcome):
    if outcome == True:
        return OUTCOME_SUCCESS
    if outcome == False:
        return OUTCOME_FAILURE
    return OUTCOME_NOT_AVAILABLE


def _serialize_steps(steps, parent_node):
    for step in steps:
        step_node = make_xml_child(parent_node, "step", "description", step.description)
        _add_time_attr(step_node, "start-time", step.start_time)
        _add_time_attr(step_node, "end-time", step.end_time)
        for entry in step.entries:
            if isinstance(entry, LogData):
                log_node = make_xml_child(step_node, "log", "level", entry.level)
                _add_time_attr(log_node, "time", entry.time)
                log_node.text = entry.message
            elif isinstance(entry, AttachmentData):
                attachment_node = make_xml_child(step_node, "attachment", "description", entry.description)
                attachment_node.text = entry.filename
            elif isinstance(entry, UrlData):
                url_node = make_xml_child(step_node, "url", "description", entry.description)
                url_node.text = entry.url
            else: # TestCheck
                check_node = make_xml_child(step_node, "check", "description", entry.description,
                                        "outcome", _serialize_outcome(entry.outcome))
                check_node.text = entry.details


def _serialize_test_data(test):
    test_node = make_xml_node("test", "name", test.name, "description", test.description,
                          "status", test.status, "status-details", test.status_details)
    _add_time_attr(test_node, "start-time", test.start_time)
    _add_time_attr(test_node, "end-time", test.end_time)
    for tag in test.tags:
        tag_node = make_xml_child(test_node, "tag")
        tag_node.text = tag
    for name, value in test.properties.items():
        property_node = make_xml_child(test_node, "property", "name", name)
        property_node.text = value
    for link in test.links:
        link_node = make_xml_child(test_node, "link", "name", link[1])
        link_node.text = link[0]
    _serialize_steps(test.steps, test_node)

    return test_node


def _serialize_hook_data(data, node):
    node.attrib["outcome"] = _serialize_outcome(data.outcome)
    _add_time_attr(node, "start-time", data.start_time)
    _add_time_attr(node, "end-time", data.end_time)
    _serialize_steps(data.steps, node)


def _serialize_suite_data(suite):
    suite_node = make_xml_node("suite", "name", suite.name, "description", suite.description)
    for tag in suite.tags:
        tag_node = make_xml_child(suite_node, "tag")
        tag_node.text = tag
    for name, value in suite.properties.items():
        property_node = make_xml_child(suite_node, "property", "name", name)
        property_node.text = value
    for link in suite.links:
        link_node = make_xml_child(suite_node, "link", "name", link[1])
        link_node.text = link[0]

    # before suite
    if suite.suite_setup:
        _serialize_hook_data(suite.suite_setup, make_xml_child(suite_node, "suite-setup"))

    # tests
    for test in suite.get_tests():
        test_node = _serialize_test_data(test)
        suite_node.append(test_node)

    # sub suites
    for sub_suite in suite.get_suites():
        sub_suite_node = _serialize_suite_data(sub_suite)
        suite_node.append(sub_suite_node)

    # after suite
    if suite.suite_teardown:
        _serialize_hook_data(suite.suite_teardown, make_xml_child(suite_node, "suite-teardown"))

    return suite_node


def serialize_report_as_tree(report):
    xml = E("lemoncheesecake-report")
    report_version_node = make_xml_child(xml, "lemoncheesecake-version")
    report_version_node.text = lemoncheesecake.__version__
    report_version_node = make_xml_child(xml, "lemoncheesecake-report-version")
    report_version_node.text = str("1.0")
    _add_time_attr(xml, "start-time", report.start_time)
    _add_time_attr(xml, "end-time", report.end_time)
    _add_time_attr(xml, "generation-time", report.report_generation_time)

    title_node = make_xml_child(xml, "title")
    title_node.text = report.title
    for name, value in report.info:
        info_node = make_xml_child(xml, "info", "name", name)
        info_node.text = value
    for name, value in report.serialize_stats():
        stat_node = make_xml_child(xml, "stat", "name", name)
        stat_node.text = value

    if report.test_session_setup:
        _serialize_hook_data(report.test_session_setup, make_xml_child(xml, "test-session-setup"))

    for suite in report.suites:
        suite_node = _serialize_suite_data(suite)
        xml.append(suite_node)

    if report.test_session_teardown:
        _serialize_hook_data(report.test_session_teardown, make_xml_child(xml, "test-session-teardown"))

    return xml


def serialize_report_as_string(report, indent_level=DEFAULT_INDENT_LEVEL):
    report = serialize_report_as_tree(report)
    indent_xml(report, indent_level=indent_level)

    if IS_PYTHON3:
        return ET.tostring(report, pretty_print=True, encoding="unicode")
    return ET.tostring(report, pretty_print=True, xml_declaration=True, encoding="utf-8")


def save_report_into_file(report, filename, indent_level=DEFAULT_INDENT_LEVEL):
    content = serialize_report_as_string(report, indent_level)
    with open(filename, "w") as fh:
        fh.write(content)


def _unserialize_datetime(value):
    return parse_timestamp(value)


def _unserialize_outcome(value):
    if value == OUTCOME_SUCCESS:
        return True
    if value == OUTCOME_FAILURE:
        return False
    if value == OUTCOME_NOT_AVAILABLE:
        return None
    raise ProgrammingError("Unknown value '%s' for outcome" % value)


def _unserialize_step_data(xml):
    step = StepData(xml.attrib["description"])
    step.start_time = _unserialize_datetime(xml.attrib["start-time"])
    step.end_time = _unserialize_datetime(xml.attrib["end-time"])
    for xml_entry in xml:
        if xml_entry.tag == "log":
            entry = LogData(xml_entry.attrib["level"], xml_entry.text, _unserialize_datetime(xml_entry.attrib["time"]))
        elif xml_entry.tag == "attachment":
            entry = AttachmentData(xml_entry.attrib["description"], xml_entry.text)
        elif xml_entry.tag == "url":
            entry = UrlData(xml_entry.attrib["description"], xml_entry.text)
        elif xml_entry.tag == "check":
            entry = CheckData(xml_entry.attrib["description"], _unserialize_outcome(xml_entry.attrib["outcome"]),
                              xml_entry.text)
        step.entries.append(entry)
    return step


def _unserialize_test_data(xml):
    test = TestData(xml.attrib["name"], xml.attrib["description"])
    test.status = xml.attrib["status"]
    test.status_details = xml.attrib.get("status-details", None)
    test.start_time = _unserialize_datetime(xml.attrib["start-time"])
    test.end_time = _unserialize_datetime(xml.attrib["end-time"])
    test.tags = [ node.text for node in xml.xpath("tag") ]
    test.properties = { node.attrib["name"]: node.text for node in xml.xpath("property") }
    test.links = [ (link.text, link.attrib["name"]) for link in xml.xpath("link") ]
    test.steps = [ _unserialize_step_data(s) for s in xml.xpath("step") ]
    return test


def _unserialize_hook_data(xml):
    data = HookData()
    data.outcome = _unserialize_outcome(xml.attrib["outcome"])
    data.start_time = _unserialize_datetime(xml.attrib["start-time"])
    data.end_time = _unserialize_datetime(xml.attrib["end-time"])
    data.steps = [ _unserialize_step_data(s) for s in xml.xpath("step") ]
    return data


def _unserialize_suite_data(xml):
    suite = SuiteData(xml.attrib["name"], xml.attrib["description"])
    suite.tags = [ node.text for node in xml.xpath("tag") ]
    suite.properties = { node.attrib["name"]: node.text for node in xml.xpath("property") }
    suite.links = [ (link.text, link.attrib["name"]) for link in xml.xpath("link") ]

    suite_setup = xml.xpath("suite-setup")
    suite_setup = suite_setup[0] if len(suite_setup) > 0 else None
    if suite_setup != None:
        suite.suite_setup = _unserialize_hook_data(suite_setup)

    for xml_test in xml.xpath("test"):
        test = _unserialize_test_data(xml_test)
        suite.add_test(test)

    suite_teardown = xml.xpath("suite-teardown")
    suite_teardown = suite_teardown[0] if len(suite_teardown) > 0 else None
    if suite_teardown != None:
        suite.suite_teardown = _unserialize_hook_data(suite_teardown)

    for xml_suite in xml.xpath("suite"):
        sub_suite = _unserialize_suite_data(xml_suite)
        suite.add_suite(sub_suite)

    return suite


def _unserialize_keyvalue_list(nodes):
    ret = [ ]
    for node in nodes:
        ret.append([node.attrib["name"], node.text])
    return ret


def load_report_from_file(filename):
    report = BoundReport()
    try:
        with open(filename, "r") as fh:
            xml = ET.parse(fh)
    except ET.LxmlError as e:
        raise InvalidReportFile(str(e))
    except IOError as e:
        raise e # re-raise as-is
    try:
        root = xml.getroot().xpath("/lemoncheesecake-report")[0]
    except IndexError:
        raise InvalidReportFile("Cannot lemoncheesecake-report element in XML")

    report.start_time = _unserialize_datetime(root.attrib["start-time"]) if "start-time" in root.attrib else None
    report.end_time = _unserialize_datetime(root.attrib["end-time"]) if "end-time" in root.attrib else None
    report.report_generation_time = _unserialize_datetime(root.attrib["generation-time"]) if "generation-time" in root.attrib else None
    report.title = root.xpath("title")[0].text
    report.info = _unserialize_keyvalue_list(root.xpath("info"))
    report.stats = _unserialize_keyvalue_list(root.xpath("stat"))

    test_session_setup = xml.xpath("test-session-setup")
    test_session_setup = test_session_setup[0] if len(test_session_setup) else None
    if test_session_setup != None:
        report.test_session_setup = _unserialize_hook_data(test_session_setup)

    for xml_suite in root.xpath("suite"):
        suite = _unserialize_suite_data(xml_suite)
        report.add_suite(suite)

    test_session_teardown = xml.xpath("test-session-teardown")
    test_session_teardown = test_session_teardown[0] if len(test_session_teardown) else None
    if test_session_teardown != None:
        report.test_session_teardown = _unserialize_hook_data(test_session_teardown)

    return report


class XmlBackend(FileReportBackend):
    name = "xml"

    def __init__(self, save_mode=SAVE_AT_EACH_FAILED_TEST):
        FileReportBackend.__init__(self, save_mode)
        self.indent_level = DEFAULT_INDENT_LEVEL

    def get_report_filename(self):
        return "report.xml"

    def is_available(self):
        return LXML_IS_AVAILABLE

    def save_report(self, filename, report):
        save_report_into_file(report, filename, self.indent_level)

    def load_report(self, filename):
        return load_report_from_file(filename).bind(self, filename)



'''
Created on Mar 18, 2016

@author: nicolas
'''

import os
import sys
import re
import inspect

IS_PYTHON3 = sys.version_info > (3,)
IS_PYTHON2 = sys.version_info < (3,)


def humanize_duration(duration, show_milliseconds=False):
    ret = ""

    if duration / 3600 >= 1:
        ret += ("%02dh" if ret else "%dh") % (duration / 3600)
        duration %= 3600

    if duration / 60 >= 1:
        ret += ("%02dm" if ret else "%dm") % (duration / 60)
        duration %= 60

    if show_milliseconds:
        if duration >= 0:
            ret += ("%06.03fs" if ret else "%.03fs") % duration
    else:
        if duration >= 1:
            ret += ("%02ds" if ret else "%ds") % duration
        if ret == "":
            ret = "%.03fs" % duration

    return ret


def get_status_color(status):
    if status == "passed":
        return "green"
    elif status == "failed":
        return "red"
    elif status == "disabled":
        return "cyan"
    else:
        return "yellow"


def object_has_method(obj, method_name):
    try:
        return callable(getattr(obj, method_name))
    except AttributeError:
        return False


def get_callable_args(cb):
    spec = inspect.getfullargspec(cb) if IS_PYTHON3 else inspect.getargspec(cb)
    args_start = 1 if inspect.ismethod(cb) else 0
    return spec.args[args_start:]


def get_resource_path(relpath):
    return os.path.join(os.path.dirname(__file__), "resources", relpath)


def get_distincts_in_list(lst):
    distincts = []
    for elem in lst:
        if elem not in distincts:
            distincts.append(elem)
    return distincts


def dict_cat(d1, d2):
    new = dict()
    new.update(d1)
    new.update(d2)
    return new


def is_string(s):
    if IS_PYTHON2:
        return type(s) in (str, unicode)
    else:
        return type(s) is str


# borrowed from https://stackoverflow.com/a/1176023
def camel_case_to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



'''
Created on Mar 29, 2016

@author: nicolas
'''

import os
import os.path as osp

from lemoncheesecake.exceptions import InvalidReportFile, ProgrammingError, method_not_implemented
from lemoncheesecake.utils import object_has_method
from lemoncheesecake.reporting.report import Report
from lemoncheesecake import events

__all__ = (
    "get_available_backends", "ReportingBackend", "ReportingSession",
    "initialize_reporting_backends",
    "save_report", "load_reports_from_dir", "load_report",
    "filter_available_reporting_backends", "filter_reporting_backends_by_capabilities",
    "CAPABILITY_REPORTING_SESSION", "CAPABILITY_SAVE_REPORT", "CAPABILITY_LOAD_REPORT"
)

CAPABILITY_REPORTING_SESSION = 0x1
CAPABILITY_SAVE_REPORT = 0x2
CAPABILITY_LOAD_REPORT = 0x4

SAVE_AT_END_OF_TESTS = 1
SAVE_AT_EACH_SUITE = 2
SAVE_AT_EACH_TEST = 3
SAVE_AT_EACH_FAILED_TEST = 4
SAVE_AT_EACH_EVENT = 5


class ReportingSession:
    pass


class ReportingBackend:
    def is_available(self):
        return True

    def get_capabilities(self):
        capabilities = 0
        if object_has_method(self, "create_reporting_session"):
            capabilities |= CAPABILITY_REPORTING_SESSION
        if object_has_method(self, "save_report"):
            capabilities |= CAPABILITY_SAVE_REPORT
        if object_has_method(self, "load_report"):
            capabilities |= CAPABILITY_LOAD_REPORT
        return capabilities

#     def create_reporting_session(self, dir, report):
#         method_not_implemented("create_reporting_session", self)
#
#     def save_report(self, filename, report):
#         method_not_implemented("serialize_report", self)
#
#     def load_report(self, filename):
#         method_not_implemented("unserialize_report", self)


def initialize_reporting_backends(backends, report_dir, report):
    for backend in backends:
        session = backend.create_reporting_session(report_dir, report)
        events.add_listener(session)


class FileReportSession(ReportingSession):
    def __init__(self, report_filename, report, save_func, save_mode):
        self.report_filename = report_filename
        self.report = report
        self.save_func = save_func
        self.save_mode = save_mode

    def save(self):
        self.save_func(self.report_filename, self.report)

    def _handle_code_end(self, is_failure):
        if (self.save_mode == SAVE_AT_EACH_TEST) or (self.save_mode == SAVE_AT_EACH_FAILED_TEST and is_failure):
            self.save()
            return

    def on_test_session_setup_end(self, event):
        self._handle_code_end(
            self.report.test_session_setup.has_failure() if self.report.test_session_setup else False
        )

    def on_test_session_teardown_end(self, event):
        self._handle_code_end(
            self.report.test_session_teardown.has_failure() if self.report.test_session_teardown else False
        )

    def on_suite_setup_end(self, event):
        suite_data = self.report.get_suite(event.suite)
        self._handle_code_end(suite_data.suite_setup and suite_data.suite_setup.has_failure())

    def on_suite_teardown_end(self, event):
        suite_data = self.report.get_suite(event.suite)
        self._handle_code_end(suite_data.suite_teardown and suite_data.suite_teardown.has_failure())

    def on_test_end(self, event):
        test_data = self.report.get_test(event.test)
        self._handle_code_end(test_data.status == "failed")

    def on_suite_end(self, event):
        if self.save_mode == SAVE_AT_EACH_SUITE:
            self.save()

    def on_log(self, event):
        if self.save_mode == SAVE_AT_EACH_EVENT:
            self.save()

    def on_check(self, event):
        if self.save_mode == SAVE_AT_EACH_EVENT:
            self.save()

    def on_test_session_end(self, event):
        self.save()


class FileReportBackend(ReportingBackend):
    def __init__(self, save_mode=SAVE_AT_EACH_FAILED_TEST):
        self.save_mode = save_mode

    def get_report_filename(self):
        method_not_implemented("get_report_filename", self)

    def create_reporting_session(self, report_dir, report):
        return FileReportSession(
            os.path.join(report_dir, self.get_report_filename()), report, self.save_report, self.save_mode
        )


def filter_available_reporting_backends(backends):
    return list(filter(lambda backend: backend.is_available(), backends))


def filter_reporting_backends_by_capabilities(backends, capabilities):
    return list(filter(lambda backend: backend.get_capabilities() & capabilities == capabilities, backends))


def get_available_backends():
    from lemoncheesecake.reporting.backends import ConsoleBackend, XmlBackend, JsonBackend, HtmlBackend, JunitBackend

    return list(filter(lambda b: b.is_available(), [ConsoleBackend(), XmlBackend(), JsonBackend(), HtmlBackend(), JunitBackend()]))


class BoundReport(Report):
    def __init__(self):
        Report.__init__(self)
        self.backend = None
        self.path = None

    def bind(self, backend, path):
        self.backend = backend
        self.path = path
        return self

    def is_bound(self):
        return self.backend is not None and self.path is not None

    def save(self):
        if not self.is_bound():
            raise ProgrammingError("Cannot save unbound report")
        save_report(self.path, self, self.backend)


def load_report_from_file(filename, backends=None):
    if backends is None:
        backends = get_available_backends()
    for backend in backends:
        if backend.get_capabilities() & CAPABILITY_LOAD_REPORT:
            try:
                return backend.load_report(filename)
            except IOError as excp:
                raise InvalidReportFile("Cannot load report from file '%s': %s" % (filename, excp))
            except InvalidReportFile:
                pass
    raise InvalidReportFile("Cannot find any suitable report backend to unserialize file '%s'" % filename)


def load_reports_from_dir(dirname, backends=None):
    for filename in [os.path.join(dirname, filename) for filename in os.listdir(dirname)]:
        if os.path.isfile(filename):
            try:
                yield load_report_from_file(filename, backends)
            except InvalidReportFile:
                pass


def load_report(path, backends=None):
    if osp.isdir(path):
        try:
            return next(load_reports_from_dir(path, backends))
        except StopIteration:
            raise InvalidReportFile("Cannot find any report in directory '%s'" % path)
    else:
        return load_report_from_file(path, backends)


def save_report(filename, report, backend):
    if not backend.get_capabilities() & CAPABILITY_SAVE_REPORT:
        raise ProgrammingError("Reporting backend '%s' does not support save operation" % backend.name)
    backend.save_report(filename, report)


'''
Created on Feb 14, 2017

@author: nicolas
'''

from __future__ import print_function

from termcolor import colored

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import get_suites_from_project
from lemoncheesecake.filter import add_run_filter_cli_args
from lemoncheesecake.project import load_project


class ShowCommand(Command):
    def get_name(self):
        return "show"
    
    def get_description(self):
        return "Show the test tree"
    
    def add_cli_args(self, cli_parser):
        add_run_filter_cli_args(cli_parser)

        group = cli_parser.add_argument_group("Display")
        group.add_argument("--no-metadata", "-i", action="store_true", help="Hide suite and test metadata")
        group.add_argument("--short", "-s", action="store_true", help="Display suite and test names instead of path")
        group.add_argument("--desc-mode", "-d", action="store_true", help="Display suite and test descriptions instead of path")
        group.add_argument("--flat-mode", "-f", action="store_true", help="Enable flat mode: display all test and suite as path without indentation nor prefix")
        self.add_color_cli_args(group)

    def get_padding(self, depth):
        return " " * (depth * self.indent)
    
    def serialize_metadata(self, obj):
        return ", ".join(
            ([colored("DISABLED", attrs=[])] if obj.disabled else []) +
            obj.tags +
            ["%s:%s" % (k, v) for k, v in obj.properties.items()] +
            [link_name or link_url for link_url, link_name in obj.links]
        )
    
    def get_test_label(self, test, suite):
        if self.show_description:
            return test.description
        if self.short:
            return test.name
        return test.path

    def get_suite_label(self, suite):
        if self.show_description:
            return suite.description
        if self.short:
            return suite.name
        return suite.path

    def show_test(self, test, suite):
        md = self.serialize_metadata(test) if self.show_metadata else ""
        if self.flat_mode:
            print("%s%s" % (self.get_test_label(test, suite), " (%s)" % md if md else ""))
        else:
            padding = self.get_padding(suite.hierarchy_depth + 1)
            test_label = self.get_test_label(test, suite)
            print("%s- %s%s" % (padding, test_label, " (%s)" % md if md else ""))
        
    def show_suite(self, suite):
        md = self.serialize_metadata(suite) if self.show_metadata else ""
        if self.flat_mode:
            print("%s%s" % (self.bold(self.get_suite_label(suite)), " (%s)" % md if md else ""))
        else:
            padding = self.get_padding(suite.hierarchy_depth)
            suite_label = self.get_suite_label(suite)
            print("%s* %s%s" % (padding, self.bold(suite_label), " (%s)" % md if md else ""))

        for test in suite.get_tests():
            self.show_test(test, suite)

        for sub_suite in suite.get_suites():
            self.show_suite(sub_suite)
    
    def show_suites(self, suites):
        for suite in suites:
            self.show_suite(suite)

    def run_cmd(self, cli_args):
        self.process_color_cli_args(cli_args)

        self.short = cli_args.short
        self.show_description = cli_args.desc_mode
        self.show_metadata = not cli_args.no_metadata
        self.flat_mode = cli_args.flat_mode
        self.indent = 4

        project = load_project()
        suites = get_suites_from_project(project, cli_args)
        
        self.show_suites(suites)
        
        return 0


'''
Created on Jan 25, 2017

@author: nicolas
'''

import argparse

from lemoncheesecake.exceptions import LemonCheesecakeException
from lemoncheesecake.cli.commands import get_commands
from lemoncheesecake.cli.utils import LEMONCHEESECAKE_VERSION


def main(args=None):
    cli_parser = argparse.ArgumentParser()
    cli_parser.add_argument("--version", "-v", action="version", version=LEMONCHEESECAKE_VERSION)
    cli_sub_parsers = cli_parser.add_subparsers(dest="command")
    for command in get_commands():
        cli_cmd_parser = cli_sub_parsers.add_parser(command.get_name(), help=command.get_description())
        try:
            command.add_cli_args(cli_cmd_parser)
        except LemonCheesecakeException as e:
            return str(e)

    cli_args = cli_parser.parse_args(args)

    command = next(cmd for cmd in get_commands() if cmd.get_name() == cli_args.command)
    try:
        return command.run_cmd(cli_args)
    except LemonCheesecakeException as e:
        return str(e)



'''
Created on Mar 26, 2016

@author: nicolas
'''

import time
import re
from decimal import Decimal

from lemoncheesecake.consts import LOG_LEVEL_ERROR, LOG_LEVEL_WARN
from lemoncheesecake.utils import humanize_duration
from lemoncheesecake.testtree import BaseTest, BaseSuite, flatten_tests, find_test, find_suite, TreeLocation

__all__ = (
    "LogData", "CheckData", "AttachmentData", "UrlData", "StepData", "TestData",
    "SuiteData", "HookData", "Report"
)

TEST_STATUSES = "passed", "failed", "skipped", "disabled"
DEFAULT_REPORT_TITLE = "Test Report"

# NB: it would be nicer to use:
# datetime.isoformat(sep=' ', timespec='milliseconds')
# unfortunately, the timespec argument is only available since Python 3.6

def format_timestamp(ts, date_time_sep=" ", skip_milliseconds=False):
    ts = round(ts, 3)
    result = time.strftime("%Y-%m-%d{sep}%H:%M:%S".format(sep=date_time_sep), time.localtime(ts))
    if not skip_milliseconds:
        result += ".%03d" % (Decimal(repr(ts)) % 1 * 1000)
    return result


def format_timestamp_as_iso_8601(ts):
    return format_timestamp(ts, date_time_sep="T", skip_milliseconds=True)


def parse_timestamp(s):
    m = re.compile("(.+)\.(\d+)").match(s)
    if not m:
        raise ValueError("s is not valid datetime representation with milliseconds precision")

    dt, milliseconds = m.group(1), int(m.group(2))

    return time.mktime(time.strptime(dt, "%Y-%m-%d %H:%M:%S")) + float(milliseconds) / 1000


def _get_duration(start_time, end_time):
    if start_time is not None and end_time is not None:
        return end_time - start_time
    else:
        return None


class LogData:
    def __init__(self, level, message, ts):
        self.level = level
        self.message = message
        self.time = ts

    def has_failure(self):
        return self.level == LOG_LEVEL_ERROR


class CheckData:
    def __init__(self, description, outcome, details=None):
        self.description = description
        self.outcome = outcome
        self.details = details

    def has_failure(self):
        return self.outcome is False


class AttachmentData:
    def __init__(self, description, filename):
        self.description = description
        self.filename = filename

    def has_failure(self):
        return False


class UrlData:
    def __init__(self, description, url):
        self.description = description
        self.url = url

    def has_failure(self):
        return False


class StepData:
    def __init__(self, description, detached=False):
        self.description = description
        self._detached = detached  # this attribute is runtime only is not intended to be serialized
        self.entries = []
        self.start_time = None
        self.end_time = None

    def has_failure(self):
        return len(list(filter(lambda entry: entry.has_failure(), self.entries))) > 0

    @property
    def duration(self):
        return _get_duration(self.start_time, self.end_time)


class TestData(BaseTest):
    def __init__(self, name, description):
        BaseTest.__init__(self, name, description)
        self.status = None
        self.status_details = None
        self.steps = []
        self.start_time = None
        self.end_time = None
        
    def has_failure(self):
        return len(list(filter(lambda step: step.has_failure(), self.steps))) > 0

    @property
    def duration(self):
        return _get_duration(self.start_time, self.end_time)


class HookData:
    def __init__(self):
        self.steps = []
        self.start_time = None
        self.end_time = None
        self.outcome = None

    def has_failure(self):
        return len(list(filter(lambda step: step.has_failure(), self.steps))) > 0

    def is_empty(self):
        return len(self.steps) == 0

    @property
    def duration(self):
        return _get_duration(self.start_time, self.end_time)


class SuiteData(BaseSuite):
    def __init__(self, name, description):
        BaseSuite.__init__(self, name, description)
        self.suite_setup = None
        self.suite_teardown = None

    @property
    def start_time(self):
        if self.suite_setup:
            return self.suite_setup.start_time
        elif len(self.get_tests()) > 0:
            return self.get_tests()[0].start_time
        else:
            return None

    @property
    def end_time(self):
        if self.suite_teardown:
            return self.suite_teardown.end_time
        elif len(self.get_tests()) > 0:
            return self.get_tests()[-1].end_time
        else:
            return None

    @property
    def duration(self):
        return _get_duration(self.start_time, self.end_time)


class _ReportStats:
    def __init__(self):
        self.tests = 0
        self.test_statuses = {s: 0 for s in TEST_STATUSES}
        self.errors = 0
        self.checks = 0
        self.check_successes = 0
        self.check_failures = 0
        self.error_logs = 0
        self.warning_logs = 0
        self.duration = 0

    def is_successful(self):
        return self.test_statuses["failed"] == 0 and self.test_statuses["skipped"] == 0

    def get_enabled_tests(self):
        return self.tests - self.test_statuses["disabled"]

    def _walk_steps(self, steps):
        for step in steps:
            for entry in step.entries:
                if isinstance(entry, CheckData):
                    self.checks += 1
                    if entry.outcome == True:
                        self.check_successes += 1
                    elif entry.outcome == False:
                        self.check_failures += 1
                if isinstance(entry, LogData):
                    if entry.level == LOG_LEVEL_WARN:
                        self.warning_logs += 1
                    elif entry.level == LOG_LEVEL_ERROR:
                        self.error_logs += 1

    def _walk_hook(self, hook):
        if hook.end_time is not None:
            self.duration += hook.end_time - hook.start_time
        if hook.has_failure():
            self.errors += 1
        self._walk_steps(hook.steps)
    walk_setup_hook = _walk_hook
    walk_teardown_hook = _walk_hook

    def walk_test(self, test):
        self.tests += 1
        if test.end_time is not None:
            self.duration += test.end_time - test.start_time
        if test.status is not None:
            self.test_statuses[test.status] += 1
        self._walk_steps(test.steps)

    def walk_suite(self, suite):
        if suite.suite_setup:
            self.walk_setup_hook(suite.suite_setup)

        if suite.suite_teardown:
            self.walk_teardown_hook(suite.suite_teardown)

        for test in suite.get_tests():
            self.walk_test(test)

        for sub_suite in suite.get_suites():
            self.walk_suite(sub_suite)

    def walk_suites(self, suites):
        for suite in suites:
            self.walk_suite(suite)


def get_stats_from_report(report):
    stats = _ReportStats()

    if report.test_session_setup:
        stats.walk_setup_hook(report.test_session_setup)

    if report.test_session_teardown:
        stats.walk_teardown_hook(report.test_session_teardown)

    stats.walk_suites(report.suites)

    return stats


def get_stats_from_suites(suites):
    stats = _ReportStats()
    stats.walk_suites(suites)

    return stats


class Report:
    def __init__(self):
        self.info = []
        self.test_session_setup = None
        self.test_session_teardown = None
        self.suites = []
        self.start_time = None
        self.end_time = None
        self.report_generation_time = None
        self.title = DEFAULT_REPORT_TITLE

    @property
    def duration(self):
        return _get_duration(self.start_time, self.end_time)

    def add_info(self, name, value):
        self.info.append([name, value])
    
    def add_suite(self, suite):
        self.suites.append(suite)
    
    def get_suites(self):
        return self.suites

    def get_suite(self, hierarchy):
        return find_suite(self.suites, hierarchy)

    def get_test(self, hierarchy):
        return find_test(self.suites, hierarchy)

    def get_stats(self):
        return get_stats_from_report(self)

    def get(self, location):
        if location.node_type == TreeLocation.TEST_SESSION_SETUP:
            return self.test_session_setup
        elif location.node_type == TreeLocation.TEST_SESSION_TEARDOWN:
            return self.test_session_teardown
        elif location.node_type == TreeLocation.SUITE_SETUP:
            suite = self.get_suite(location.node_hierarchy)
            return suite.suite_setup
        elif location.node_type == TreeLocation.SUITE_TEARDOWN:
            suite = self.get_suite(location.node_hierarchy)
            return suite.suite_teardown
        elif location.node_type == TreeLocation.TEST:
            return self.get_test(location.node_hierarchy)
        else:
            raise Exception("Unknown location type %s" % location.node_type)

    def serialize_stats(self):
        stats = self.get_stats()
        enabled_tests = stats.get_enabled_tests()
        return (
            ("Start time", time.asctime(time.localtime(self.start_time))),
            ("End time", time.asctime(time.localtime(self.end_time)) if self.end_time else "n/a"),
            ("Duration", humanize_duration(self.end_time - self.start_time) if self.end_time else "n/a"),
            ("Tests", str(stats.tests)),
            ("Successful tests", str(stats.test_statuses["passed"])),
            ("Successful tests in %", "%d%%" % (float(stats.test_statuses["passed"]) / enabled_tests * 100 if enabled_tests else 0)),
            ("Failed tests", str(stats.test_statuses["failed"])),
            ("Skipped tests", str(stats.test_statuses["skipped"])),
            ("Disabled tests", str(stats.test_statuses["disabled"])),
            ("Errors", str(stats.errors))
        )


def flatten_steps(suites):
    for test in flatten_tests(suites):
        for step in test.steps:
            yield step



'''
Created on Jan 24, 2016

@author: nicolas
'''

import os
import sys
import traceback

from lemoncheesecake.utils import IS_PYTHON3, get_distincts_in_list
from lemoncheesecake.runtime import *
from lemoncheesecake.runtime import initialize_runtime, set_runtime_location
from lemoncheesecake.reporting import Report, initialize_report_writer, initialize_reporting_backends
from lemoncheesecake.exceptions import AbortTest, AbortSuite, AbortAllTests, FixtureError, \
    UserError, serialize_current_exception
from lemoncheesecake import events
from lemoncheesecake.testtree import TreeLocation


def _get_fixtures_used_in_suite(suite):
    fixtures = suite.get_fixtures()

    for test in suite.get_tests():
        fixtures.extend(test.get_fixtures())

    return get_distincts_in_list(fixtures)


def _get_fixtures_used_in_suite_recursively(suite):
    fixtures = _get_fixtures_used_in_suite(suite)

    for sub_suite in suite.get_suites():
        fixtures.extend(_get_fixtures_used_in_suite_recursively(sub_suite))

    return get_distincts_in_list(fixtures)


class _Runner:
    def __init__(self, suites, fixture_registry, reporting_backends, report_dir, stop_on_failure=False):
        self.suites = suites
        self.fixture_registry = fixture_registry
        self.reporting_backends = reporting_backends
        self.report_dir = report_dir
        self.stop_on_failure = stop_on_failure
        # attributes set at runtime:
        self._report = None
        self._session = None
        self._abort_all_tests = False
        self._abort_suite = None

    def _is_suite_setup_failure(self, suite):
        suite_data = self._report.get_suite(suite)
        return suite_data.suite_setup and suite_data.suite_setup.has_failure()

    def _is_suite_teardown_failure(self, suite):
        suite_data = self._report.get_suite(suite)
        return suite_data.suite_teardown and suite_data.suite_teardown.has_failure()

    def _is_test_failure(self, test):
        test_data = self._report.get_test(test)
        return test_data.has_failure()

    def get_fixtures_with_dependencies_for_scope(self, direct_fixtures, scope):
        fixtures = []
        for fixture in direct_fixtures:
            fixtures.extend(self.fixture_registry.get_fixture_dependencies(fixture))
        fixtures.extend(direct_fixtures)
        return [f for f in get_distincts_in_list(fixtures) if self.fixture_registry.get_fixture_scope(f) == scope]

    def get_fixtures_to_be_executed_for_session_prerun(self):
        fixtures = []
        for suite in self.suites:
            fixtures.extend(_get_fixtures_used_in_suite_recursively(suite))
        return self.get_fixtures_with_dependencies_for_scope(get_distincts_in_list(fixtures), "session_prerun")

    def get_fixtures_to_be_executed_for_session(self):
        fixtures = []
        for suite in self.suites:
            fixtures.extend(_get_fixtures_used_in_suite_recursively(suite))
        return self.get_fixtures_with_dependencies_for_scope(get_distincts_in_list(fixtures), "session")

    def get_fixtures_to_be_executed_for_suite(self, suite):
        return self.get_fixtures_with_dependencies_for_scope(_get_fixtures_used_in_suite(suite), "suite")

    def get_fixtures_to_be_executed_for_test(self, test):
        return self.get_fixtures_with_dependencies_for_scope(test.get_fixtures(), "test")

    def run_setup_funcs(self, funcs, failure_checker):
        teardown_funcs = []
        for setup_func, teardown_func in funcs:
            if setup_func:
                try:
                    setup_func()
                except (Exception, KeyboardInterrupt) as e:
                    self.handle_exception(e)
                    break
                else:
                    if failure_checker():
                        break
                    else:
                        teardown_funcs.append(teardown_func)
            else:
                teardown_funcs.append(teardown_func)
        return teardown_funcs

    def run_teardown_funcs(self, teardown_funcs):
        count = 0
        for teardown_func in teardown_funcs:
            if teardown_func:
                try:
                    teardown_func()
                except (Exception, KeyboardInterrupt) as e:
                    self.handle_exception(e)
                else:
                    count += 1
            else:
                count += 1
        return count

    def get_fixture_as_funcs(self, fixture):
        return [
            lambda: self.fixture_registry.execute_fixture(fixture),
            lambda: self.fixture_registry.teardown_fixture(fixture)
        ]

    def get_setup_suite_as_func(self, suite):
        setup_suite = suite.get_hook("setup_suite")
        if setup_suite == None:
            return None

        fixtures_names = suite.get_hook_params("setup_suite")
        def func():
            fixtures = self.fixture_registry.get_fixture_results(fixtures_names)
            setup_suite(**fixtures)

        return func

    def inject_fixtures_into_suite(self, suite):
        fixture_names = suite.get_injected_fixture_names()
        fixtures = self.fixture_registry.get_fixture_results(fixture_names)
        suite.inject_fixtures(fixtures)

    def handle_exception(self, excp, suite=None):
        if isinstance(excp, AbortTest):
            log_error(str(excp))
        elif isinstance(excp, AbortSuite):
            log_error(str(excp))
            self._abort_suite = suite
        elif isinstance(excp, AbortAllTests):
            log_error(str(excp))
            self._abort_all_tests = True
        elif isinstance(excp, KeyboardInterrupt):
            log_error("All tests have been interrupted manually by the user")
            self._abort_all_tests = True
        else:
            # FIXME: use exception instead of last implicit stacktrace
            stacktrace = traceback.format_exc()
            if not IS_PYTHON3:
                stacktrace = stacktrace.decode("utf-8", "replace")
            log_error("Caught unexpected exception while running test: " + stacktrace)

    def _begin_test(self, test):
        events.fire(events.TestStartEvent(test))
        set_runtime_location(TreeLocation.in_test(test))

    def _end_test(self, test):
        events.fire(events.TestEndEvent(test))

    def _begin_test_setup(self, test):
        events.fire(events.TestSetupStartEvent(test))
        set_step("Setup test")

    def _end_test_setup(self, test, outcome):
        events.fire(events.TestSetupEndEvent(test, outcome))

    def _begin_test_teardown(self, test):
        events.fire(events.TestTeardownStartEvent(test))
        set_step("Teardown test")

    def _end_test_teardown(self, test, outcome):
        events.fire(events.TestTeardownEndEvent(test, outcome))

    def run_test(self, test, suite):
        ###
        # Checker whether the test must be executed or not
        ###
        if test.is_disabled():
            events.fire(events.TestDisabledEvent(test, ""))
            return

        if self._abort_suite:
            events.fire(events.TestSkippedEvent(test, "Cannot execute this test: the tests of this test suite have been aborted."))
            return

        if self._abort_all_tests:
            events.fire(events.TestSkippedEvent(test, "Cannot execute this test: all tests have been aborted."))
            return

        ###
        # Begin test
        ###

        self._begin_test(test)

        ###
        # Setup test (setup and fixtures)
        ###
        test_setup_error = False
        setup_teardown_funcs = []
        teardown_funcs = []

        setup_teardown_funcs.append([
            (lambda: suite.get_hook("setup_test")(test.name)) if suite.has_hook("setup_test") else None,
            (lambda: suite.get_hook("teardown_test")(test.name)) if suite.has_hook("teardown_test") else None
        ])
        setup_teardown_funcs.extend([
            self.get_fixture_as_funcs(f) for f in self.get_fixtures_to_be_executed_for_test(test)
        ])

        if len(list(filter(lambda p: p[0] is not None, setup_teardown_funcs))) > 0:
            self._begin_test_setup(test)
            teardown_funcs = self.run_setup_funcs(
                setup_teardown_funcs, lambda: self._is_test_failure(test)
            )
            if len(teardown_funcs) != len(setup_teardown_funcs):
                test_setup_error = True
            self._end_test_setup(test, not test_setup_error)
        else:
            teardown_funcs = [p[1] for p in setup_teardown_funcs if p[1] is not None]
        
        if self.stop_on_failure and test_setup_error:
            self._abort_all_tests = True

        ###
        # Run test:
        ###
        if not test_setup_error:
            test_func_params = self.fixture_registry.get_fixture_results(test.get_fixtures())
            set_step(test.description)
            try:
                test.callback(**test_func_params)
            except (Exception, KeyboardInterrupt) as e:
                self.handle_exception(e, suite)

        ###
        # Teardown
        ###
        if len(list(filter(lambda f: f is not None, teardown_funcs))) > 0:
            self._begin_test_teardown(test)
            self.run_teardown_funcs(teardown_funcs)
            self._end_test_teardown(test, not self._is_test_failure(test))

        if self.stop_on_failure and self._is_test_failure(test):
            self._abort_all_tests = True

        self._end_test(test)

    def _begin_suite(self, suite):
        events.fire(events.SuiteStartEvent(suite))

    def _end_suite(self, suite):
        events.fire(events.SuiteEndEvent(suite))

    def _begin_suite_setup(self, suite):
        events.fire(events.SuiteSetupStartEvent(suite))
        set_runtime_location(TreeLocation.in_suite_setup(suite))
        set_step("Setup suite")

    def _end_suite_setup(self, suite):
        events.fire(events.SuiteSetupEndEvent(suite))

    def _begin_suite_teardown(self, suite):
        events.fire(events.SuiteTeardownStartEvent(suite))
        set_runtime_location(TreeLocation.in_suite_teardown(suite))
        set_step("Teardown suite")

    def _end_suite_teardown(self, suite):
        events.fire(events.SuiteTeardownEndEvent(suite))

    def run_suite(self, suite):
        ###
        # Begin suite
        ###
        self._begin_suite(suite)

        ###
        # Setup suite (suites and fixtures)
        ###
        teardown_funcs = []
        if not self._abort_all_tests:
            setup_teardown_funcs = []
            # first, fixtures must be executed
            setup_teardown_funcs.extend([
                self.get_fixture_as_funcs(f) for f in self.get_fixtures_to_be_executed_for_suite(suite)
            ])
            # then, fixtures must be injected into suite
            setup_teardown_funcs.append((lambda: self.inject_fixtures_into_suite(suite), None))
            # and at the end, the setup_suite hook of the suite will be called
            setup_teardown_funcs.append([
                self.get_setup_suite_as_func(suite), suite.get_hook("teardown_suite")
            ])

            if len(list(filter(lambda p: p[0] is not None, setup_teardown_funcs))) > 0:
                self._begin_suite_setup(suite)
                teardown_funcs = self.run_setup_funcs(
                    setup_teardown_funcs, lambda: self._is_suite_setup_failure(suite)
                )
                if len(teardown_funcs) != len(setup_teardown_funcs):
                    self._abort_suite = suite
                self._end_suite_setup(suite)
            else:
                teardown_funcs = [p[1] for p in setup_teardown_funcs if p[1] is not None]
            
            if self.stop_on_failure and self._abort_suite:
                self._abort_all_tests = True

        ###
        # Run tests
        ###
        for test in suite.get_tests():
            self.run_test(test, suite)

        ###
        # Teardown suite
        ###
        if len(list(filter(lambda f: f is not None, teardown_funcs))) > 0:
            self._begin_suite_teardown(suite)
            self.run_teardown_funcs(teardown_funcs)
            if self.stop_on_failure and self._is_suite_teardown_failure(suite):
                self._abort_all_tests = True
            self._end_suite_teardown(suite)

        # reset the abort suite flag
        if self._abort_suite:
            self._abort_suite = None

        ###
        # Run sub suites
        ###
        for sub_suite in suite.get_suites():
            self.run_suite(sub_suite)

        ###
        # End of suite
        ###

        self._end_suite(suite)

    def _begin_test_session(self):
        events.fire(events.TestSessionStartEvent(self._report))

    def _end_test_session(self):
        events.fire(events.TestSessionEndEvent(self._report))

    def _begin_test_session_setup(self):
        events.fire(events.TestSessionSetupStartEvent())
        set_runtime_location(TreeLocation.in_test_session_setup())
        set_step("Setup test session")

    def _end_test_session_setup(self):
        events.fire(events.TestSessionSetupEndEvent())

    def _begin_test_session_teardown(self):
        events.fire(events.TestSessionTeardownStartEvent())
        set_runtime_location(TreeLocation.in_test_session_teardown())
        set_step("Teardown test session")

    def _end_test_session_teardown(self):
        events.fire(events.TestSessionTeardownEndEvent())

    def run_session(self):
        # initialize runtime & global test variables
        self._report = Report()
        self._report.add_info("Command line", " ".join([os.path.basename(sys.argv[0])] + sys.argv[1:]))
        self._session = initialize_report_writer(self._report)
        initialize_runtime(self.report_dir, self._report, self.fixture_registry)
        initialize_reporting_backends(self.reporting_backends, self.report_dir, self._report)
        self._abort_all_tests = False
        self._abort_suite = None

        self._begin_test_session()

        # setup test session
        setup_teardown_funcs = []
        teardown_funcs = []
        setup_teardown_funcs.extend([
            self.get_fixture_as_funcs(f) for f in self.get_fixtures_to_be_executed_for_session()
        ])

        if len(list(filter(lambda p: p[0] is not None, setup_teardown_funcs))) > 0:
            self._begin_test_session_setup()
            teardown_funcs = self.run_setup_funcs(
                setup_teardown_funcs, lambda: self._session.report.test_session_setup.has_failure()
            )
            if len(teardown_funcs) != len(setup_teardown_funcs):
                self._abort_all_tests = True
            self._end_test_session_setup()
        else:
            teardown_funcs = [p[1] for p in setup_teardown_funcs if p[1] is not None]

        # run suites
        for suite in self.suites:
            self.run_suite(suite)

        # teardown_test_session handling
        if len(list(filter(lambda f: f is not None, teardown_funcs))) > 0:
            self._begin_test_session_teardown()
            self.run_teardown_funcs(teardown_funcs)
            self._end_test_session_teardown()

        self._end_test_session()

    def run(self):
        executed_fixtures = []

        # setup pre_session fixtures
        errors = []
        for fixture in self.get_fixtures_to_be_executed_for_session_prerun():
            try:
                self.fixture_registry.execute_fixture(fixture)
            except UserError:
                raise
            except (Exception, KeyboardInterrupt):
                errors.append("Got the following exception when executing fixture '%s' (scope 'session_prerun')%s" % (
                    fixture, serialize_current_exception(show_stacktrace=True)
                ))
                break
            executed_fixtures.append(fixture)

        if not errors:
            self.run_session()

        # teardown pre_session fixtures
        for fixture in executed_fixtures:
            try:
                self.fixture_registry.teardown_fixture(fixture)
            except UserError:
                raise
            except (Exception, KeyboardInterrupt):
                errors.append("Got the following exception on fixture '%s' teardown (scope 'session_prerun')%s" % (
                    fixture, serialize_current_exception(show_stacktrace=True)
                ))

        if errors:
            raise FixtureError("\n".join(errors))

        stats = self._report.get_stats()
        return stats.is_successful()


def run_suites(suites, fixture_registry, reporting_backends, report_dir, stop_on_failure=False):
    """
    Run suites.

    - suites: a list of already loaded suites (see lemoncheesecake.loader.load_suites)
    - reporting_backends: instance of reporting backends that will be used to report test results
    - report_dir: an existing directory where report files will be stored
    """
    runner = _Runner(suites, fixture_registry, reporting_backends, report_dir, stop_on_failure)
    return runner.run()  # TODO: return a Report instance instead



'''
Created on Mar 28, 2017

@author: nicolas
'''

from lemoncheesecake.matching.base import Matcher, match_success, match_failure, match_result, \
    got, got_value, to_be, merge_match_result_descriptions

__all__ = (
    "all_of", "any_of", "anything", "something", "existing", "present", "is_", "not_"
)


def _make_item(content, prefix="- "):
    lines = content.split("\n")
    return "\n".join(
        ["    " + prefix + lines[0]] +
        ["      " + line for line in lines[1:]]
    )


def _make_items(items, prefix="- ", relationship="and"):
    return [
        _make_item(item, prefix=prefix + relationship + " " if i > 0 else prefix)
            for i, item in enumerate(items)
    ]


def _serialize_sub_matcher_result(matcher, result):
    content = "%s => %s" % (matcher.short_description(conjugate=True), "OK" if result.is_success() else "KO")
    if result.description is not None:
        content += ", %s" % result.description
    return content


class AllOf(Matcher):
    def __init__(self, matchers):
        self.matchers = matchers

    def short_description(self, conjugate=False):
        return ":"

    def description(self, conjugate=False):
        return "\n".join(
            [":"] +
            [
                _make_item(matcher.description(conjugate=conjugate), prefix="- and " if i > 0 else "- ")
                    for i, matcher in enumerate(self.matchers)
            ]
        )

    def matches(self, actual):
        results = []
        is_success = True
        for matcher in self.matchers:
            result = matcher.matches(actual)
            results.append((matcher, result))
            if result.is_failure():
                is_success = False
                break

        match_details = "\n".join(
            [got() + ":"] +
            _make_items([
                _serialize_sub_matcher_result(matcher, result) for matcher, result in results
            ], relationship="and")
        )

        return match_result(is_success, match_details)


def all_of(*matchers):
    """Test if all matchers match (logical AND between matchers)."""
    return AllOf(map(is_, matchers))


class AnyOf(Matcher):
    def __init__(self, matchers):
        self.matchers = matchers

    def short_description(self, conjugate=False):
        return ":"

    def description(self, conjugate=False):
        return "\n".join(
            [":"] +
            [
                _make_item(matcher.description(conjugate=conjugate), prefix="- or " if i > 0 else "- ")
                    for i, matcher in enumerate(self.matchers)
            ]
        )

    def matches(self, actual):
        results = []
        for matcher in self.matchers:
            match = matcher.matches(actual)
            if match.is_success():
                return match
            results.append(match)

        return match_failure(merge_match_result_descriptions(results))


def any_of(*matchers):
    """Test if at least one of the matcher match (logical OR between matchers)"""
    return AnyOf(map(is_, matchers))


class Anything(Matcher):
    def __init__(self, wording="anything"):
        self.wording = wording

    def description(self, conjugate=False):
        return "%s %s" % (to_be(conjugate), self.wording)

    def matches(self, actual):
        return match_success(got_value(actual))


def anything():
    """Matches anything (always succeed, whatever the actual value)"""
    return Anything()


def something():
    """Same thing as the 'anything' matcher but use 'something' in the matcher description"""
    return Anything(wording="something")


def existing():
    """Same thing as the 'anything' matcher but use 'existing' in the matcher description"""
    return Anything(wording="existing")


def present():
    """Same thing as the 'anything' matcher but use 'present' in the matcher description"""
    return Anything(wording="present")


def is_(matcher):
    """If the function argument is not an instance of Matcher, wrap it into
    a matcher using equal_to, otherwise return the matcher argument as-is"""
    from lemoncheesecake.matching.matchers.value import equal_to
    return matcher if isinstance(matcher, Matcher) else equal_to(matcher)


class Not(Matcher):
    def __init__(self, matcher):
        self.matcher = matcher

    def description(self, conjugate=False):
        return "not %s" % self.matcher.description(conjugate)

    def matches(self, actual):
        result = self.matcher.matches(actual)
        return match_result(not result.outcome, result.description)


def not_(matcher):
    """Negates the matcher in argument"""
    return Not(is_(matcher))



'''
Created on Mar 27, 2017

@author: nicolas
'''

from lemoncheesecake.matching.base import (
    MatchExpected, Matcher, match_result, match_success, match_failure,
    got_value, serialize_value, to_be, to_have
)
from lemoncheesecake.matching.matchers.composites import is_
from lemoncheesecake.matching.matchers.types_ import is_bool

__all__ = (
    "equal_to", "not_equal_to", "greater_than", "greater_than_or_equal_to",
    "less_than", "less_than_or_equal_to",
    "is_between", "is_none", "is_not_none", "has_length",
    "is_true", "is_false"
)


class EqualTo(MatchExpected):
    def description(self, conjugate=False):
        return "%s equal to %s" % (to_be(conjugate), serialize_value(self.expected))

    def matches(self, actual):
        from lemoncheesecake.matching import DISPLAY_DETAILS_WHEN_EQUAL

        if actual == self.expected:
            return match_success(got_value(actual)) if DISPLAY_DETAILS_WHEN_EQUAL else match_success()
        else:
            return match_failure(got_value(actual))


def equal_to(expected):
    """Test if value is equal to expected"""
    return EqualTo(expected)


def _comparator(comparison_description, comparison_func):
    def wrapper(expected):
        class _Comparator(MatchExpected):
            def description(self, conjugate=False):
                return "%s %s %s" % (to_be(conjugate), comparison_description, serialize_value(self.expected))

            def matches(self, actual):
                return match_result(comparison_func(actual, self.expected), got_value(actual))

        return _Comparator(expected)

    wrapper.__doc__ = """Test if value is %s expected""" % comparison_description
    return wrapper


not_equal_to = _comparator("not equal to", lambda a, e: a != e)

greater_than = _comparator("greater than", lambda a, e: a > e)
greater_than_or_equal_to = _comparator("greater than or equal to", lambda a, e: a >= e)

less_than = _comparator("less than", lambda a, e: a < e)
less_than_or_equal_to = _comparator("greater than or equal to", lambda a, e: a <= e)


class IsBetween(Matcher):
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def description(self, conjugate=False):
        return "%s between %s and %s" % (to_be(conjugate), self.min, self.max)

    def matches(self, actual):
        return match_result(self.min <= actual <= self.max, got_value(actual))


def is_between(min, max):
    """Test if value is between min and max"""
    return IsBetween(min, max)


def is_none():
    """Test if value is None"""
    return equal_to(None)


def is_not_none():
    """Test if value is not None"""
    return not_equal_to(None)


class HasLength(Matcher):
    def __init__(self, matcher):
        self.matcher = matcher

    def description(self, conjugate=False):
        return "%s a length that %s" % (to_have(conjugate), self.matcher.description(conjugate=True))

    def matches(self, actual):
        return self.matcher.matches(len(actual))


def has_length(length):
    """Test if value has a length of"""
    return HasLength(is_(length))


def is_true():
    """Test if value is true (boolean type)"""
    return is_bool(True)


def is_false():
    """Test if value is false (boolean type)"""
    return is_bool(False)



'''
Created on Jun 14, 2017

@author: nicolas
'''

from decimal import Decimal

# This junit backend implementation is based on the documentation 
# http://llg.cubic.org/docs/junit/
# also see:
# https://confluence.atlassian.com/display/BAMBOO/JUnit+parsing+in+Bamboo

try:
    from lxml import etree as ET
    from lxml.builder import E
    LXML_IS_AVAILABLE = True
except ImportError:
    LXML_IS_AVAILABLE = False

from lemoncheesecake.reporting.backend import FileReportBackend, SAVE_AT_EACH_FAILED_TEST
from lemoncheesecake.reporting.report import LogData, CheckData, format_timestamp_as_iso_8601
from lemoncheesecake.utils import IS_PYTHON3
from lemoncheesecake.consts import LOG_LEVEL_ERROR
from lemoncheesecake.reporting.backends.xml import make_xml_child, make_xml_node, indent_xml, set_node_attr, \
    DEFAULT_INDENT_LEVEL

def format_duration(duration):
    return "%d.%03d" % (int(duration), Decimal(round(duration, 3)) % 1 * 1000)

def _serialize_test_data(test):
    junit_test = make_xml_node(
        "testcase",
        "name", test.name,
        "time", format_duration(test.end_time - test.start_time),
    )
    
    if test.status == "skipped":
        make_xml_child(junit_test, "skipped")
    else:
        for step in test.steps:
            for step_entry in step.entries:
                if isinstance(step_entry, CheckData) and step_entry.outcome == False:
                    make_xml_child(junit_test, "failure", "message", "failed check in step '%s'" % step.description)
                elif isinstance(step_entry, LogData) and step_entry.level == LOG_LEVEL_ERROR:
                    make_xml_child(junit_test, "error", "message", "error log in step '%s'" % step.description)

    return junit_test

def _serialize_suite_data(suite):
    junit_testsuites = []
    tests = suite.get_tests()
    if tests:
        junit_testsuite = make_xml_node(
            "testsuite", 
            "name", suite.path,
            "tests", str(len(tests)),
            "failures", str(len(list(filter(lambda test: test.status == "failed", tests)))),
            "skipped", str(len(list(filter(lambda test: test.status == "skipped", tests)))),
            "time", format_duration(tests[-1].end_time - tests[0].start_time),
            "timestamp", format_timestamp_as_iso_8601(tests[0].start_time)
        )
        junit_testsuites.append(junit_testsuite)
        for test in tests:
            junit_test = _serialize_test_data(test)
            junit_testsuite.append(junit_test)
    
    for sub_suite in suite.get_suites():
        junit_sub_testsuites = _serialize_suite_data(sub_suite)
        for junit_sub_testsuite in junit_sub_testsuites:
            junit_testsuites.append(junit_sub_testsuite)
    
    return junit_testsuites
    
def serialize_report_as_tree(report):
    junit_report_testsuites = E("testsuites")
    
    report_stats = report.get_stats()
    set_node_attr(junit_report_testsuites, "tests", str(report_stats.test_statuses["passed"]))
    set_node_attr(junit_report_testsuites, "failures", str(report_stats.test_statuses["failed"]))
    if report.end_time != None:
        set_node_attr(junit_report_testsuites, "time", format_duration(report.end_time - report.start_time))
    
    for suite in report.suites:
        junit_testsuites = _serialize_suite_data(suite)
        for junit_testsuite in junit_testsuites:
            junit_report_testsuites.append(junit_testsuite)

    return junit_report_testsuites

def serialize_report_as_string(report, indent_level=DEFAULT_INDENT_LEVEL):
    report = serialize_report_as_tree(report)
    indent_xml(report, indent_level=indent_level)

    if IS_PYTHON3:
        return ET.tostring(report, pretty_print=True, encoding="unicode")
    return ET.tostring(report, pretty_print=True, xml_declaration=True, encoding="utf-8")

def save_report_into_file(report, filename, indent_level=DEFAULT_INDENT_LEVEL):
    content = serialize_report_as_string(report, indent_level)
    with open(filename, "w") as fh:
        fh.write(content)

class JunitBackend(FileReportBackend):
    name = "junit"

    def __init__(self, save_mode=SAVE_AT_EACH_FAILED_TEST):
        FileReportBackend.__init__(self, save_mode)
        self.indent_level = DEFAULT_INDENT_LEVEL

    def get_report_filename(self):
        return "report-junit.xml"

    def is_available(self):
        return LXML_IS_AVAILABLE

    def save_report(self, filename, report):
        save_report_into_file(report, filename, self.indent_level)



from __future__ import print_function

import os
import sys
import time

try:
    import slacker
    SLACKER_IS_AVAILABLE = True
except ImportError:
    SLACKER_IS_AVAILABLE = False

from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession
from lemoncheesecake.exceptions import UserError
from lemoncheesecake.utils import humanize_duration


class BaseSlackReportingSession(ReportingSession):
    def __init__(self, auth_token, channel, proxy=None):
        self.slack = slacker.Slacker(
            auth_token, http_proxy=proxy, https_proxy=proxy
        )
        self.channel = channel
        self.errors = []

    def send_message(self, message):
        try:
            self.slack.chat.post_message(self.channel, message)
        except slacker.Error as excp:
            self.errors.append("Error while notifying Slack channel/user '%s', got: %s" % (
                self.channel, excp
            ))

    def show_errors(self):
        if len(self.errors) == 0:
            return

        print("Slack reporting backend, got the following errors while sending messages:", file=sys.stderr)
        for error in self.errors:
            print("- %s" % error, file=sys.stderr)


def percent(val, of):
    return "%d%%" % ((float(val) / of * 100) if of else 0)


def get_message_template_parameters():
    return {
        "start_time": lambda report, stats: time.asctime(time.localtime(report.start_time)),
        "end_time": lambda report, stats: time.asctime(time.localtime(report.end_time)),
        "duration": lambda report, stats: humanize_duration(report.end_time - report.start_time),

        "total": lambda report, stats: stats.tests,
        "enabled": lambda report, stats: stats.get_enabled_tests(),

        "passed": lambda report, stats: stats.test_statuses["passed"],
        "passed_pct": lambda report, stats: percent(stats.test_statuses["passed"], of=stats.get_enabled_tests()),

        "failed": lambda report, stats: stats.test_statuses["failed"],
        "failed_pct": lambda report, stats: percent(stats.test_statuses["failed"], of=stats.get_enabled_tests()),

        "skipped": lambda report, stats: stats.test_statuses["skipped"],
        "skipped_pct": lambda report, stats: percent(stats.test_statuses["skipped"], of=stats.get_enabled_tests()),

        "disabled": lambda report, stats: stats.test_statuses["disabled"],
        "disabled_pct": lambda report, stats: percent(stats.test_statuses["disabled"], of=stats.tests)
    }


def build_message_parameters(report, stats):
    return {name: func(report, stats) for name, func in get_message_template_parameters().items()}


def build_message_empty_parameters():
    return {name: "" for name in get_message_template_parameters()}


def check_message_template_validity(template):
    try:
        template.format(**build_message_empty_parameters())
    except KeyError as excp:
        raise UserError("Invalid Slack message template, unknown variable: %s" % excp)


class EndOfTestsNotifier(BaseSlackReportingSession):
    def __init__(self, auth_token, channel, message_template, proxy=None, only_notify_failure=False):
        BaseSlackReportingSession.__init__(self, auth_token, channel, proxy=proxy)
        check_message_template_validity(message_template)
        self.message_template = message_template
        self.only_notify_failure = only_notify_failure

    def build_message_parameters(self, report, stats):
        enabled_tests = stats.get_enabled_tests()

        def percent(val, of):
            return "%d%%" % ((float(val) / of * 100) if of else 0)

        return {
            "start_time": time.asctime(time.localtime(report.start_time)),
            "end_time": time.asctime(time.localtime(report.end_time)),
            "duration": humanize_duration(report.end_time - report.start_time),

            "total": stats.tests,
            "enabled": enabled_tests,

            "passed": stats.test_statuses["passed"],
            "passed_pct": percent(stats.test_statuses["passed"], of=enabled_tests),

            "failed": stats.test_statuses["failed"],
            "failed_pct": percent(stats.test_statuses["failed"], of=enabled_tests),

            "skipped": stats.test_statuses["skipped"],
            "skipped_pct": percent(stats.test_statuses["skipped"], of=enabled_tests),

            "disabled": stats.test_statuses["disabled"],
            "disabled_pct": percent(stats.test_statuses["disabled"], of=stats.tests)
        }

    def on_test_session_end(self, event):
        stats = event.report.get_stats()
        if self.only_notify_failure and stats.is_successful():
            return

        message = self.message_template.format(**build_message_parameters(event.report, stats))
        self.send_message(message)

        self.show_errors()


def get_env_var(name, optional=False, default=None):
    try:
        return os.environ[name]
    except KeyError as excp:
        if optional:
            return default
        else:
            raise UserError("Slack reporting backend, cannot get environment variable %s" % excp)


def get_slack_auth_token():
    return get_env_var("SLACK_AUTH_TOKEN")


def get_slack_http_proxy():
    return get_env_var("SLACK_HTTP_PROXY", optional=True, default=None)


def get_slack_channel():
    return get_env_var("SLACK_CHANNEL")


def get_message_template():
    return get_env_var(
        "SLACK_MESSAGE_TEMPLATE", optional=True, default="{passed} passed, {failed} failed, {skipped} skipped"
    )


def get_only_notify_on_failure():
    return "SLACK_ONLY_NOTIFY_FAILURE" in os.environ


class SlackReportingBackend(ReportingBackend):
    name = "slack"

    def is_available(self):
        return SLACKER_IS_AVAILABLE

    def create_reporting_session(self, report_dir, report):
        return EndOfTestsNotifier(
            get_slack_auth_token(), get_slack_channel(), get_message_template(),
            proxy=get_slack_http_proxy(), only_notify_failure=get_only_notify_on_failure()
        )



import time
import re
import inspect

from lemoncheesecake.utils import camel_case_to_snake_case


def _get_event_name_from_class_name(class_name):
    return re.sub("_event$", "", camel_case_to_snake_case(class_name))


class Event(object):
    def __init__(self):
        self.time = time.time()

    @classmethod
    def get_name(cls):
        return _get_event_name_from_class_name(cls.__name__)


class EventType:
    def __init__(self, event_class):
        self._event_class = event_class
        self._handlers = []

    def subscribe(self, handler):
        self._handlers.append(handler)

    def unsubscribe(self, handler):
        self._handlers.remove(handler)

    def reset(self):
        self._handlers = []

    def fire(self, event):
        for handler in self._handlers:
            handler(event)


def _get_event_name(val):
    return val.get_name() if inspect.isclass(val) and issubclass(val, Event) else val


class EventManager:
    def __init__(self):
        self._event_types = {}

    def register_events(self, *event_classes):
        for event_class in event_classes:
            self._event_types[event_class.get_name()] = EventType(event_class)

    def reset(self, event_name=None):
        if event_name is None:
            for event_type in self._event_types.values():
                event_type.reset()
        else:
            self._event_types[event_name].reset()

    def subscribe_to_event(self, event, handler):
        self._event_types[_get_event_name(event)].subscribe(handler)

    def subscribe_to_events(self, event_handler_pairs):
        for event, handler in event_handler_pairs.items():
            self.subscribe_to_event(event, handler)

    def add_listener(self, listener):
        for event_name in self._event_types:
            handler_name = "on_%s" % event_name
            handler = getattr(listener, handler_name, None)
            if handler and callable(handler):
                self.subscribe_to_event(event_name, handler)

    def unsubscribe_from_event(self, event, handler):
        self._event_types[_get_event_name(event)].unsubscribe(handler)

    def fire(self, event):
        self._event_types[event.__class__.get_name()].fire(event)


eventmgr = EventManager()
register_event = eventmgr.register_events
register_events = eventmgr.register_events
subscribe_to_event = eventmgr.subscribe_to_event
subscribe_to_events = eventmgr.subscribe_to_events
unsubscribe_from_event = eventmgr.unsubscribe_from_event
add_listener = eventmgr.add_listener
reset = eventmgr.reset
fire = eventmgr.fire


def event(class_):
    register_event(class_)
    return class_


###
# Events related to the test session
###

class _ReportEvent(Event):
    def __init__(self, report):
        super(_ReportEvent, self).__init__()
        self.report = report


@event
class TestSessionStartEvent(_ReportEvent):
    pass


@event
class TestSessionEndEvent(_ReportEvent):
    pass


@event
class TestSessionSetupStartEvent(Event):
    pass


@event
class TestSessionSetupEndEvent(Event):
    pass


@event
class TestSessionTeardownStartEvent(Event):
    pass


@event
class TestSessionTeardownEndEvent(Event):
    pass


###
# Suite events
###

class _SuiteEvent(Event):
    def __init__(self, suite):
        super(_SuiteEvent, self).__init__()
        self.suite = suite


@event
class SuiteStartEvent(_SuiteEvent):
    pass


@event
class SuiteEndEvent(_SuiteEvent):
    pass


@event
class SuiteSetupStartEvent(_SuiteEvent):
    pass


@event
class SuiteSetupEndEvent(_SuiteEvent):
    pass


@event
class SuiteTeardownStartEvent(_SuiteEvent):
    pass


@event
class SuiteTeardownEndEvent(_SuiteEvent):
    pass


###
# Test events
###

class _TestEvent(Event):
    def __init__(self, test):
        super(_TestEvent, self).__init__()
        self.test = test


@event
class TestStartEvent(_TestEvent):
    pass


@event
class TestEndEvent(_TestEvent):
    pass


@event
class TestSkippedEvent(_TestEvent):
    def __init__(self, test, reason):
        super(TestSkippedEvent, self).__init__(test)
        self.skipped_reason = reason


@event
class TestDisabledEvent(_TestEvent):
    def __init__(self, test, reason):
        super(TestDisabledEvent, self).__init__(test)
        self.disabled_reason = reason


@event
class TestSetupStartEvent(_TestEvent):
    pass


@event
class TestSetupEndEvent(_TestEvent):
    def __init__(self, test, outcome):
        super(TestSetupEndEvent, self).__init__(test)
        self.setup_outcome = outcome


@event
class TestTeardownStartEvent(_TestEvent):
    pass


@event
class TestTeardownEndEvent(_TestEvent):
    def __init__(self, test, outcome):
        super(TestTeardownEndEvent, self).__init__(test)
        self.teardown_outcome = outcome


###
# Transverse test execution events
###

class RuntimeEvent(Event):
    def __init__(self, location):
        super(RuntimeEvent, self).__init__()
        self.location = location


@event
class StepEvent(RuntimeEvent):
    def __init__(self, location, description, detached=False):
        super(StepEvent, self).__init__(location)
        self.step_description = description
        self.detached = detached


@event
class StepEndEvent(RuntimeEvent):
    def __init__(self, location, step):
        super(StepEndEvent, self).__init__(location)
        self.step = step


class SteppedEvent(RuntimeEvent):
    def __init__(self, location, step):
        super(SteppedEvent, self).__init__(location)
        self.step = step


@event
class LogEvent(SteppedEvent):
    def __init__(self, location, step, level, message):
        super(LogEvent, self).__init__(location, step)
        self.log_level = level
        self.log_message = message


@event
class CheckEvent(SteppedEvent):
    def __init__(self, location, step, description, outcome, details=None):
        super(CheckEvent, self).__init__(location, step)
        self.check_description = description
        self.check_outcome = outcome
        self.check_details = details


@event
class LogAttachmentEvent(SteppedEvent):
    def __init__(self, location, step, path, filename, description):
        super(LogAttachmentEvent, self).__init__(location, step)
        self.attachment_path = path
        self.attachment_filename = filename
        self.attachment_description = description


@event
class LogUrlEvent(SteppedEvent):
    def __init__(self, location, step, url, description):
        super(LogUrlEvent, self).__init__(location, step)
        self.url = url
        self.url_description = description



'''
Created on Mar 12, 2017

@author: nicolas
'''

from __future__ import print_function

from termcolor import colored
from terminaltables import AsciiTable


def bold(s):
    return colored(s, attrs=["bold"])


def print_table(title, headers, lines):
    if lines:
        print("%s:" % title)
        print(AsciiTable([headers] + lines).table)
    else:
        print("%s: <none>" % title)
    print()



'''
Created on Apr 4, 2017

@author: nicolas
'''

import re

from lemoncheesecake.matching.base import MatchExpected, match_result, got_value

__all__ = (
    "starts_with", "ends_with", "contains_string", "match_pattern"
)

_REGEXP_TYPE = type(re.compile("dummy"))


class StartsWith(MatchExpected):
    def description(self, conjugate=False):
        return '%s with "%s"' % ("starts" if conjugate else "to start", self.expected)

    def matches(self, actual):
        return match_result(actual.startswith(self.expected), got_value(actual))


def starts_with(s):
    """Test if string begins with given prefix"""
    return StartsWith(s)


class EndsWith(MatchExpected):
    def description(self, conjugate=False):
        return '%s with "%s"' % ("ends" if conjugate else "to end", self.expected)

    def matches(self, actual):
        return match_result(actual.endswith(self.expected), got_value(actual))


def ends_with(s):
    """Test if string ends with given suffix"""
    return EndsWith(s)


class ContainsString(MatchExpected):
    def description(self, conjugate=False):
        return '%s "%s"' % ("contains" if conjugate else "to contain", self.expected)

    def matches(self, actual):
        return match_result(self.expected in actual, got_value(actual))


def contains_string(s):
    """Test if string contains sub string"""
    return ContainsString(s)


class MatchPattern(MatchExpected):
    def description(self, conjugate=False):
        return '%s "%s"' % ("matches pattern" if conjugate else "to match pattern", self.expected.pattern)

    def matches(self, actual):
        return match_result(self.expected.search(actual) != None, got_value(actual))


def match_pattern(pattern):
    """Test if string matches given pattern (using the `search` method of the `re` module)"""
    return MatchPattern(pattern if type(pattern) == _REGEXP_TYPE else re.compile(pattern))



'''
Created on Mar 27, 2016

@author: nicolas
'''

import re
import json
from collections import OrderedDict

import lemoncheesecake
from lemoncheesecake.reporting.backend import BoundReport, FileReportBackend, SAVE_AT_EACH_FAILED_TEST
from lemoncheesecake.reporting.report import (
    LogData, CheckData, AttachmentData, UrlData, StepData, TestData, HookData, SuiteData,
    format_timestamp, parse_timestamp
)
from lemoncheesecake.exceptions import InvalidReportFile

JS_PREFIX = "var reporting_data = "


def _serialize_time(ts):
    return format_timestamp(ts) if ts else None


def _dict(*args):
    d = OrderedDict()
    i = 0
    while i < len(args):
        d[args[i]] = args[i+1]
        i += 2
    return d


def _serialize_steps(steps):
    json_steps = []
    for step in steps:
        json_step = _dict(
            "description", step.description,
            "start_time", _serialize_time(step.start_time),
            "end_time", _serialize_time(step.end_time),
            "entries", []
        )
        json_steps.append(json_step)
        for entry in step.entries:
            if isinstance(entry, LogData):
                entry = _dict("type", "log", "level", entry.level, "message", entry.message, "time", _serialize_time(entry.time))
            elif isinstance(entry, AttachmentData):
                entry = _dict("type", "attachment", "description", entry.description, "filename", entry.filename)
            elif isinstance(entry, UrlData):
                entry = _dict("type", "url", "description", entry.description, "url", entry.url)
            else: # TestCheck
                entry = _dict("type", "check", "description", entry.description, "outcome", entry.outcome, "details", entry.details)
            json_step["entries"].append(entry)
    return json_steps


def _serialize_common_data(obj):
    return _dict(
        "name", obj.name, "description", obj.description,
        "tags", obj.tags,
        "properties", obj.properties,
        "links", [ _dict("name", link[1], "url", link[0]) for link in obj.links ]
    )


def _serialize_test_data(test):
    serialized = _serialize_common_data(test)
    serialized.update(_dict(
        "start_time", _serialize_time(test.start_time),
        "end_time", _serialize_time(test.end_time),
        "steps", _serialize_steps(test.steps),
        "status", test.status,
        "status_details", test.status_details
    ))
    return serialized


def _serialize_hook_data(hook_data):
    return _dict(
        "start_time", _serialize_time(hook_data.start_time),
        "end_time", _serialize_time(hook_data.end_time),
        "steps", _serialize_steps(hook_data.steps),
        "outcome", hook_data.outcome
    )


def _serialize_suite_data(suite):
    json_suite = _serialize_common_data(suite)
    json_suite.update(_dict(
        "tests", [ _serialize_test_data(t) for t in suite.get_tests() ],
        "suites", [ _serialize_suite_data(s) for s in suite.get_suites() ]
    ))
    if suite.suite_setup:
        json_suite["suite_setup"] = _serialize_hook_data(suite.suite_setup)
    if suite.suite_teardown:
        json_suite["suite_teardown"] = _serialize_hook_data(suite.suite_teardown)

    return json_suite


def serialize_report_into_json(report):
    serialized = _dict(
        "lemoncheesecake_version", lemoncheesecake.__version__,
        "lemoncheesecake_report_version", 1.0,
        "start_time", _serialize_time(report.start_time),
        "end_time", _serialize_time(report.end_time),
        "generation_time", _serialize_time(report.report_generation_time),
        "title", report.title,
        "info", [ [ n, v ] for n, v in report.info ],
        "stats", [ [ n, v ] for n, v in report.serialize_stats() ]
    )

    if report.test_session_setup:
        serialized["test_session_setup"] = _serialize_hook_data(report.test_session_setup)

    serialized["suites"] = [ _serialize_suite_data(s) for s in report.suites ]

    if report.test_session_teardown:
        serialized["test_session_teardown"] = _serialize_hook_data(report.test_session_teardown)

    return serialized


def save_report_into_file(data, filename, javascript_compatibility=True, pretty_formatting=False):
    report = serialize_report_into_json(data)
    file = open(filename, "w")
    if javascript_compatibility:
        file.write(JS_PREFIX)
    if pretty_formatting:
        file.write(json.dumps(report, indent=4))
    else:
        file.write(json.dumps(report))
    file.close()


def _unserialize_time(t):
    return parse_timestamp(t)


def _unserialize_step_data(js):
    step = StepData(js["description"])
    step.start_time = _unserialize_time(js["start_time"])
    step.end_time = _unserialize_time(js["end_time"])
    for js_entry in js["entries"]:
        if js_entry["type"] == "log":
            entry = LogData(js_entry["level"], js_entry["message"], _unserialize_time(js_entry["time"]))
        elif js_entry["type"] == "attachment":
            entry = AttachmentData(js_entry["description"], js_entry["filename"])
        elif js_entry["type"] == "url":
            entry = UrlData(js_entry["description"], js_entry["url"])
        elif js_entry["type"] == "check":
            entry = CheckData(js_entry["description"], js_entry["outcome"], js_entry["details"])
        step.entries.append(entry)
    return step


def _unserialize_test_data(js):
    test = TestData(js["name"], js["description"])
    test.status = js["status"]
    test.status_details = js["status_details"]
    test.start_time = _unserialize_time(js["start_time"])
    test.end_time = _unserialize_time(js["end_time"])
    test.tags = js["tags"]
    test.properties = js["properties"]
    test.links = [ (link["url"], link["name"]) for link in js["links"] ]
    test.steps = [ _unserialize_step_data(s) for s in js["steps"] ]
    return test


def _unserialize_hook_data(js):
    data = HookData()
    data.outcome = js["outcome"]
    data.start_time = _unserialize_time(js["start_time"])
    data.end_time = _unserialize_time(js["end_time"])
    data.steps = [ _unserialize_step_data(s) for s in js["steps"] ]

    return data


def _unserialize_suite_data(js):
    suite = SuiteData(js["name"], js["description"])
    suite.tags = js["tags"]
    suite.properties = js["properties"]
    suite.links = [(link["url"], link["name"]) for link in js["links"]]

    if "suite_setup" in js:
        suite.suite_setup = _unserialize_hook_data(js["suite_setup"])

    for js_test in js["tests"]:
        test = _unserialize_test_data(js_test)
        suite.add_test(test)

    if "suite_teardown" in js:
        suite.suite_teardown = _unserialize_hook_data(js["suite_teardown"])

    for js_suite in js["suites"]:
        sub_suite = _unserialize_suite_data(js_suite)
        suite.add_suite(sub_suite)

    return suite


def load_report_from_file(filename):
    report = BoundReport()
    try:
        with open(filename, "r") as fh:
            js_content = fh.read()
    except IOError as e:
        raise e # re-raise as-is

    js_content = re.sub("^" + JS_PREFIX, "", js_content)

    try:
        js = json.loads(js_content)
    except ValueError as e:
        raise InvalidReportFile(str(e))

    if "lemoncheesecake_report_version" not in js:
        raise InvalidReportFile("Cannot find 'lemoncheesecake_report_version' in JSON")

    report.title = js["title"]
    report.info = js["info"]
    report.stats = js["stats"]
    report.start_time = _unserialize_time(js["start_time"])
    report.end_time = _unserialize_time(js["end_time"])
    report.report_generation_time = _unserialize_time(js["generation_time"])

    if "test_session_setup" in js:
        report.test_session_setup = _unserialize_hook_data(js["test_session_setup"])

    for js_suite in js["suites"]:
        suite = _unserialize_suite_data(js_suite)
        report.add_suite(suite)

    if "test_session_teardown" in js:
        report.test_session_teardown = _unserialize_hook_data(js["test_session_teardown"])

    return report


class JsonBackend(FileReportBackend):
    name = "json"

    def __init__(self, save_mode=SAVE_AT_EACH_FAILED_TEST, javascript_compatibility=True, pretty_formatting=False):
        FileReportBackend.__init__(self, save_mode)
        self.javascript_compatibility = javascript_compatibility
        self.pretty_formatting = pretty_formatting

    def get_report_filename(self):
        return "report.js"

    def save_report(self, filename, report):
        save_report_into_file(
            report, filename,
            javascript_compatibility=self.javascript_compatibility, pretty_formatting=self.pretty_formatting
        )

    def load_report(self, filename):
        return load_report_from_file(filename).bind(self, filename)


'''
Created on Mar 19, 2016

@author: nicolas
'''

import os
import os.path as p
from shutil import copy, copytree

from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession


class HtmlBackend(ReportingBackend):
    name = "html"

    def __init__(self, offline_mode=False):
        self.offline_mode = offline_mode

    def create_reporting_session(self, report_dir, report):
        class WriteHtmlReport(ReportingSession):
            def __init__(self, report_dir, offline_mode):
                self.report_dir = report_dir
                self.offline_mode = offline_mode

            def on_test_session_start(self, event):
                html_resource_dir = p.join(p.dirname(__file__), p.pardir, p.pardir, "resources", "html")
                report_resource_dir = p.join(self.report_dir, ".html")

                os.mkdir(report_resource_dir)
                copy(p.join(html_resource_dir, "report.js"), report_resource_dir)
                copy(p.join(html_resource_dir, "report.css"), report_resource_dir)

                if self.offline_mode:
                    copy(p.join(html_resource_dir, "report_offline.html"), p.join(self.report_dir, "report.html"))
                    copy(p.join(html_resource_dir, "jquery-1.12.3.min.js"), report_resource_dir)
                    copytree(p.join(html_resource_dir, "bootstrap-3.3.6-dist"), p.join(report_resource_dir, "bootstrap-3.3.6-dist"))
                    copy(p.join(html_resource_dir, "bootstrap-slate.min.css"),
                         p.join(report_resource_dir, "bootstrap-3.3.6-dist", "css", "bootstrap.min.css"))
                else:
                    copy(p.join(html_resource_dir, "report.html"), self.report_dir)

        return WriteHtmlReport(report_dir, self.offline_mode)



'''
Created on Sep 8, 2016

@author: nicolas
'''

from lemoncheesecake.exceptions import InvalidMetadataError, ProgrammingError, InternalError
from lemoncheesecake.utils import get_distincts_in_list, get_callable_args
from lemoncheesecake.testtree import BaseTest, BaseSuite

SUITE_HOOKS = "setup_test", "teardown_test", "setup_suite", "teardown_suite"

__all__ = (
    "Test", "Suite"
)


class Test(BaseTest):
    def __init__(self, name, description, callback):
        BaseTest.__init__(self, name, description)
        self.callback = callback
        self.disabled = False

    def is_disabled(self):
        node = self
        while node is not None:
            if node.disabled:
                return True
            node = node.parent_suite
        return False

    def get_fixtures(self):
        return get_callable_args(self.callback)


def _assert_valid_hook_name(hook_name):
    if hook_name not in SUITE_HOOKS:
        raise InternalError("Invalid hook name '%s'" % hook_name)


class InjectedFixture:
    def __init__(self, fixture_name):
        self.fixture_name = fixture_name


def _load_injected_fixtures(obj):
    fixtures = {}
    for attr_name in dir(obj):
        if attr_name.startswith("__"):
            continue
        sym = getattr(obj, attr_name)
        if isinstance(sym, InjectedFixture):
            fixtures[sym.fixture_name or attr_name] = attr_name
    return fixtures


class Suite(BaseSuite):
    def __init__(self, obj, name, description):
        BaseSuite.__init__(self, name, description)
        self.obj = obj
        self.rank = 0
        self._hooks = {}
        self._injected_fixtures = _load_injected_fixtures(obj)
        self.disabled = False

    def add_hook(self, hook_name, func):
        _assert_valid_hook_name(hook_name)
        self._hooks[hook_name] = func

    def has_hook(self, hook_name):
        _assert_valid_hook_name(hook_name)
        return hook_name in self._hooks

    def get_hook(self, hook_name):
        _assert_valid_hook_name(hook_name)
        return self._hooks.get(hook_name)

    def get_hook_params(self, hook_name):
        hook = self.get_hook(hook_name)
        assert hook != None
        return get_callable_args(hook)

    def get_injected_fixture_names(self):
        return self._injected_fixtures.keys()

    def inject_fixtures(self, fixtures):
        for fixture_name, fixture_value in fixtures.items():
            attr_name = self._injected_fixtures[fixture_name]
            setattr(self.obj, attr_name, fixture_value)

    def assert_test_is_unique_in_suite(self, test):
        try:
            next(t for t in self._tests if t.description == test.description)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A test with description '%s' is already registered in test suite %s" % (test.description, self.path)
            )

        try:
            next(t for t in self._tests if t.name == test.name)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A test with name '%s' is already registered in test suite %s" % (test.name, self.path)
            )

    def assert_sub_suite_is_unique_in_suite(self, sub_suite):
        try:
            next(s for s in self._suites if s.description == sub_suite.description)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A sub test suite with description '%s' is already registered in test suite %s" % (sub_suite.name, self.path)
            )

        try:
            next(s for s in self._suites if s.name == sub_suite.name)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A sub test suite with name '%s' is already registered in test suite %s" % (sub_suite.name, self.path)
            )

    def add_test(self, test):
        self.assert_test_is_unique_in_suite(test)
        BaseSuite.add_test(self, test)

    def add_suite(self, suite):
        self.assert_sub_suite_is_unique_in_suite(suite)
        BaseSuite.add_suite(self, suite)

    def get_fixtures(self):
        fixtures = []

        fixtures.extend(self._injected_fixtures.keys())

        suite_setup = self.get_hook("setup_suite")
        if suite_setup:
            fixtures.extend(get_callable_args(suite_setup))

        return get_distincts_in_list(fixtures)



'''
Created on Feb 14, 2017

@author: nicolas
'''

from __future__ import print_function
from functools import reduce

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.display import print_table
from lemoncheesecake.cli.utils import get_suites_from_project
from lemoncheesecake.filter import add_run_filter_cli_args
from lemoncheesecake.testtree import flatten_suites
from lemoncheesecake.project import load_project


class Stats:
    def __init__(self):
        self.tests_nb = 0
        self.disabled_tests_nb = 0
        self.suites_nb = 0
        self.non_empty_suites_nb = 0
        self.tags = {}
        self.properties = {}
        self.links = {}


def compute_stats(suites):
    stats = Stats()

    for suite in flatten_suites(suites):
        stats.suites_nb += 1
        for test in suite.get_tests():
            stats.tests_nb += 1
            if test.is_disabled():
                stats.disabled_tests_nb += 1
            for tag in test.hierarchy_tags:
                stats.tags[tag] = stats.tags.get(tag, 0) + 1
            for prop, value in test.hierarchy_properties.items():
                if prop not in stats.properties:
                    stats.properties[prop] = {}
                if value not in stats.properties[prop]:
                    stats.properties[prop][value] = 0
                stats.properties[prop][value] += 1
            for link in test.hierarchy_links:
                stats.links[link] = stats.links.get(link, 0) + 1
        else:
            stats.non_empty_suites_nb += 1

    return stats


class StatsCommand(Command):
    def get_name(self):
        return "stats"

    def get_description(self):
        return "Display statistics about the project's tests"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Display")
        self.add_color_cli_args(group)
        add_run_filter_cli_args(cli_parser)

    def run_cmd(self, cli_args):
        self.process_color_cli_args(cli_args)

        project = load_project()
        suites = get_suites_from_project(project, cli_args)

        stats = compute_stats(suites)

        def percent_of_tests(val):
            return "%2d%%" % (float(val) / stats.tests_nb * 100)

        # Show tags
        lines = []
        for tag in sorted(stats.tags.keys(), key=lambda k: stats.tags[k], reverse=True):
            lines.append([self.bold(tag), stats.tags[tag], percent_of_tests(stats.tags[tag])])
        print_table(self.bold("Tags"), ["Tag", "Tests", "In %"], lines)

        # Show properties
        lines = []
        prop_names = sorted(
            stats.properties.keys(),
            key=lambda k: reduce(lambda x, y: x + y, stats.properties[k].values(), 0),
            reverse=True
        )
        for prop_name in prop_names:
            prop_values = sorted(
                stats.properties[prop_name].keys(),
                key=lambda k: stats.properties[prop_name][k],
                reverse=True
            )
            for prop_value in prop_values:
                lines.append([
                    self.bold(prop_name), self.bold(prop_value),
                    stats.properties[prop_name][prop_value],
                    percent_of_tests(stats.properties[prop_name][prop_value])
                ])
        print_table(self.bold("Properties"), ["Property", "Value", "Tests", "In %"], lines)

        # Show links
        lines = []
        for link in sorted(stats.links.keys(), key=lambda k: stats.links[k], reverse=True):
            lines.append([
                self.bold(link[1] or "-"), link[0], stats.links[link], percent_of_tests(stats.links[link])
            ])
        print_table(self.bold("Links"), ["Name", "URL", "Tests", "In %"], lines)

        tests_info = self.bold("%d tests" % stats.tests_nb)
        if stats.disabled_tests_nb > 0:
            tests_info += " (among which %s disabled tests)" % stats.disabled_tests_nb

        suites_info = self.bold("%d suites" % stats.non_empty_suites_nb)
        if stats.suites_nb > stats.non_empty_suites_nb:
            suites_info += " (+ %d empty suites)" % (stats.suites_nb - stats.non_empty_suites_nb)

        print("Total: %s in %s" % (tests_info, suites_info))
        print()

        return 0



'''
Created on Jan 24, 2016

@author: nicolas
'''

from lemoncheesecake.reporting.report import *
from lemoncheesecake import events
from lemoncheesecake.exceptions import ProgrammingError

__all__ = "ReportWriter", "initialize_report_writer"


class ReportWriter:
    def __init__(self, report):
        self.report = report

    def _get_test_data(self, test):
        return self.report.get_test(test)

    def _get_suite_data(self, suite):
        return self.report.get_suite(suite)

    def _add_step_entry(self, entry, event):
        report_node_data = self.report.get(event.location)
        step = self._lookup_step(report_node_data.steps, event.step)
        if step.end_time:
            raise ProgrammingError("Cannot update step '%s', it is already ended" % step.description)
        step.entries.append(entry)

    @staticmethod
    def _start_hook(ts):
        hook_data = HookData()
        hook_data.start_time = ts
        return hook_data

    @staticmethod
    def _end_hook(hook_data, ts):
        if hook_data:
            hook_data.end_time = ts
            hook_data.outcome = not hook_data.has_failure()

    @staticmethod
    def _lookup_step(steps, step):
        if step is None:
            return steps[-1]
        else:
            try:
                return next(s for s in reversed(steps) if s.description == step)
            except StopIteration:
                raise ProgrammingError("Cannot find step '%s'" % step)

    @staticmethod
    def _lookup_current_step(steps):
        for step in reversed(steps):
            if not step._detached:
                return step
        return None

    @staticmethod
    def _finalize_steps(steps, end_time):
        if not steps:
            return
        for step in steps[:]:
            if step.entries:
                # set step end time (last step or detached step)
                if not step.end_time:
                    step.end_time = end_time
            else:
                # remove empty step
                steps.remove(step)

    def on_test_session_start(self, event):
        self.report.start_time = event.time

    def on_test_session_end(self, event):
        self.report.end_time = event.time
        self.report.report_generation_time = self.report.end_time

    def on_test_session_setup_start(self, event):
        self.report.test_session_setup = self._start_hook(event.time)

    def on_test_session_setup_end(self, event):
        self._finalize_steps(self.report.test_session_setup.steps, event.time)

        if self.report.test_session_setup.is_empty():
            self.report.test_session_setup = None
        else:
            self._end_hook(self.report.test_session_setup, event.time)

    def on_test_session_teardown_start(self, event):
        self.report.test_session_teardown = self._start_hook(event.time)

    def on_test_session_teardown_end(self, event):
        self._finalize_steps(self.report.test_session_teardown.steps, event.time)

        if self.report.test_session_teardown.is_empty():
            self.report.test_session_teardown = None
        else:
            self._end_hook(self.report.test_session_teardown, event.time)

    def on_suite_start(self, event):
        suite = event.suite
        suite_data = SuiteData(suite.name, suite.description)
        suite_data.tags.extend(suite.tags)
        suite_data.properties.update(suite.properties)
        suite_data.links.extend(suite.links)
        if suite.parent_suite:
            parent_suite_data = self._get_suite_data(suite.parent_suite)
            parent_suite_data.add_suite(suite_data)
        else:
            self.report.add_suite(suite_data)

    def on_suite_setup_start(self, event):
        suite_data = self._get_suite_data(event.suite)
        suite_data.suite_setup = self._start_hook(event.time)

    def on_suite_setup_end(self, event):
        suite_data = self._get_suite_data(event.suite)
        self._finalize_steps(suite_data.suite_setup.steps, event.time)

        if suite_data.suite_setup.is_empty():
            suite_data.suite_setup = None
        else:
            self._end_hook(suite_data.suite_setup, event.time)

    def on_suite_teardown_start(self, event):
        suite_data = self._get_suite_data(event.suite)

        suite_data.suite_teardown = self._start_hook(event.time)

    def on_suite_teardown_end(self, event):
        suite_data = self._get_suite_data(event.suite)
        self._finalize_steps(suite_data.suite_teardown.steps, event.time)

        if suite_data.suite_teardown.is_empty():
            suite_data.suite_teardown = None
        else:
            self._end_hook(suite_data.suite_teardown, event.time)

    def on_test_start(self, event):
        test = event.test

        test_data = TestData(test.name, test.description)
        test_data.tags.extend(test.tags)
        test_data.properties.update(test.properties)
        test_data.links.extend(test.links)
        test_data.start_time = event.time

        suite_data = self._get_suite_data(event.test.parent_suite)
        suite_data.add_test(test_data)

    def on_test_end(self, event):
        test_data = self._get_test_data(event.test)
        self._finalize_steps(test_data.steps, event.time)

        test_data.status = "failed" if test_data.has_failure() else "passed"
        test_data.end_time = event.time

    def _bypass_test(self, test, status, status_details, time):
        test_data = TestData(test.name, test.description)
        test_data.tags.extend(test.tags)
        test_data.properties.update(test.properties)
        test_data.links.extend(test.links)
        test_data.end_time = test_data.start_time = time
        test_data.status = status
        test_data.status_details = status_details

        suite_data = self._get_suite_data(test.parent_suite)
        suite_data.add_test(test_data)

    def on_test_skipped(self, event):
        self._bypass_test(event.test, "skipped", event.skipped_reason, event.time)

    def on_test_disabled(self, event):
        self._bypass_test(event.test, "disabled", "", event.time)

    def on_step(self, event):
        report_node_data = self.report.get(event.location)
        current_step = self._lookup_current_step(report_node_data.steps)
        if current_step:
            current_step.end_time = event.time

        new_step = StepData(event.step_description, detached=event.detached)
        new_step.start_time = event.time
        report_node_data.steps.append(new_step)

    def on_step_end(self, event):
        report_node_data = self.report.get(event.location)
        step = self._lookup_step(report_node_data.steps, event.step)
        if not step._detached:
            raise ProgrammingError("Cannot end step '%s', only detached steps can be explicitly ended" % step.description)
        step.end_time = event.time

    def on_log(self, event):
        self._add_step_entry(
            LogData(event.log_level, event.log_message, event.time), event
        )

    def on_check(self, event):
        self._add_step_entry(
            CheckData(event.check_description, event.check_outcome, event.check_details), event
        )

    def on_log_attachment(self, event):
        self._add_step_entry(
            AttachmentData(event.attachment_description, event.attachment_path), event
        )

    def on_log_url(self, event):
        self._add_step_entry(
            UrlData(event.url_description, event.url), event
        )


def initialize_report_writer(report):
    writer = ReportWriter(report)
    events.add_listener(writer)
    return writer



'''
Created on Sep 8, 2016

@author: nicolas
'''

import sys
import traceback


class LemonCheesecakeException(Exception):
    message_prefix = None

    def __str__(self):
        s = Exception.__str__(self)
        if self.message_prefix:
            s = "%s, %s" % (self.message_prefix, s)
        return s

class LemonCheesecakeInternalError(LemonCheesecakeException):
    pass
InternalError = LemonCheesecakeInternalError


class ProgrammingError(LemonCheesecakeException):
    pass


class ProjectError(LemonCheesecakeException):
    pass


class MethodNotImplemented(ProgrammingError):
    def __init__(self, obj, method_name):
        ProgrammingError.__init__(self, "Class '%s' must implement the method '%s'" % (obj.__class__.__name__, method_name))


def method_not_implemented(method_name, obj):
    raise MethodNotImplemented(obj, method_name)


class ModuleImportError(LemonCheesecakeException):
    pass


class FixtureError(LemonCheesecakeException):
    pass


class InvalidSuiteError(LemonCheesecakeException):
    pass


class InvalidMetadataError(ProgrammingError):
    pass


class VisibilityConditionNotMet(LemonCheesecakeException):
    pass


class UnknownReportBackendError(LemonCheesecakeException):
    pass


class AbortTest(LemonCheesecakeException):
    message_prefix = "The test has been aborted"

    def __init__(self, reason):
        LemonCheesecakeException.__init__(self, reason)


class AbortSuite(LemonCheesecakeException):
    message_prefix = "The suite has been aborted"

    def __init__(self, reason):
        LemonCheesecakeException.__init__(self, reason)


class AbortAllTests(LemonCheesecakeException):
    message_prefix = "All tests have been aborted"

    def __init__(self, reason):
        LemonCheesecakeException.__init__(self, reason)


class UserError(LemonCheesecakeException):
    pass


class InvalidReportFile(LemonCheesecakeException):
    pass


class CannotFindTreeNode(LemonCheesecakeException):
    pass


def serialize_current_exception(show_stacktrace=True):
    if show_stacktrace:
        return "\n" + "<" * 72 + "\n" + traceback.format_exc() + ">" * 72
    else:
        return " " + str(sys.exc_info()[1])



from .run import RunCommand
from .bootstrap import BootstrapCommand
from .show import ShowCommand
from .fixtures import FixturesCommand
from .stats import StatsCommand
from .report import ReportCommand
from .diff import DiffCommand
from .version import VersionCommand
from .top import TopTests, TopSuites, TopSteps


def get_commands():
    return [
        RunCommand(), BootstrapCommand(),
        ShowCommand(), FixturesCommand(), StatsCommand(),
        ReportCommand(), DiffCommand(),
        TopTests(), TopSuites(), TopSteps(),
        VersionCommand()
    ]



"""
This module gather public API to be used within suites, this module can be imported
as this for convenience:
import lemoncheesecake.api as lcc
"""

from lemoncheesecake.suite import Test, add_test_into_suite, add_test_in_suite, add_tests_in_suite, \
    get_metadata, suite, test, tags, prop, link, disabled, conditional, hidden, inject_fixture
from lemoncheesecake.runtime import *
from lemoncheesecake.checkers import *
from lemoncheesecake.matching import *
from lemoncheesecake.fixtures import fixture
from lemoncheesecake.exceptions import AbortTest, AbortSuite, AbortAllTests, UserError

# for pydoc
__all__ = dir()



'''
Created on Dec 10, 2016

@author: nicolas
'''

import os
import os.path as osp
import shutil

from lemoncheesecake.suite import load_suites_from_directory
from lemoncheesecake.fixtures import load_fixtures_from_directory
from lemoncheesecake.validators import MetadataPolicy
from lemoncheesecake.reporting import ConsoleBackend, HtmlBackend, JsonBackend, XmlBackend, JunitBackend, \
    ReportPortalBackend, SlackReportingBackend, filter_available_reporting_backends
from lemoncheesecake.reporting.reportdir import create_report_dir_with_rotation
from lemoncheesecake.exceptions import ProjectError, UserError, serialize_current_exception
from lemoncheesecake.utils import get_resource_path
from lemoncheesecake.importer import import_module

PROJECT_CONFIG_FILE = "project.py"


class HasCustomCliArgs:
    """"Mixin class for project configuration with custom CLI args"""
    def add_custom_cli_args(self, cli_parser):
        """Setup custom CLI args in cli_parser"""
        pass


class HasPreRunHook:
    """Mixin class for project configuration that requires code to be executed before tests"""
    def pre_run(self, report_dir):
        """Pre-run hook"""
        pass


class HasPostRunHook:
    """Mixin class for project configuration that requires code to be executed after tests"""
    def post_run(self, report_dir):
        """Post-run hook"""
        pass


class HasMetadataPolicy:
    """Mixin class project configuration that a metadata policy"""
    def get_metadata_policy(self):
        """Return a `lemoncheesecake.validators.MetadataPolicy` instance"""
        return MetadataPolicy()


class ProjectConfiguration:
    """Abstract class for project configuration"""

    def get_suites(self):
        """Return a list of `lemoncheesecake.suite.Suite` instances"""
        raise NotImplementedError()
    
    def get_fixtures(self):
        """Return a list of `lemoncheesecake.fixtures.Fixture` instances"""
        return []
    
    def get_all_reporting_backends(self):
        """Return a list of `lemoncheesecake.reporting.ReportingBackend` instances supported by the project"""
        raise NotImplementedError()
    
    def get_default_reporting_backends_for_test_run(self):
        """Return a list of `lemoncheesecake.reporting.ReportingBackend` that will used by default by lcc run"""
        raise NotImplementedError()

    def create_report_dir(self, top_dir):
        """Create a new report directory within `top_dir` (in case the user does not provide a custom report directory)"""
        raise NotImplementedError()

    def get_report_title(self):
        """Return the report title"""
        return None

    def get_report_info(self):
        """Return a list of key/value tuple to be added to report info"""
        return []


class SimpleProjectConfiguration(ProjectConfiguration):
    def __init__(self, suites_dir, fixtures_dir=None, report_title=None):
        self._suites_dir = suites_dir
        self._fixtures_dir = fixtures_dir
        self._report_title = report_title
        self.console_backend = ConsoleBackend()
        self.json_backend = JsonBackend()
        self.xml_backend = XmlBackend()
        self.junit_backend = JunitBackend()
        self.html_backend = HtmlBackend()
        self.reportportal_backend = ReportPortalBackend()
        self.slack_backend = SlackReportingBackend()
    
    def get_suites(self):
        return load_suites_from_directory(self._suites_dir)
    
    def get_fixtures(self):
        return load_fixtures_from_directory(self._fixtures_dir) if self._fixtures_dir else []

    def get_report_title(self):
        return self._report_title
    
    def get_all_reporting_backends(self):
        return [
            self.console_backend, self.json_backend, self.html_backend,
            self.xml_backend, self.junit_backend, self.reportportal_backend,
            self.slack_backend
        ]

    def get_default_reporting_backends_for_test_run(self):
        return [self.console_backend, self.json_backend, self.html_backend]
    
    def create_report_dir(self, top_dir):
        return create_report_dir_with_rotation(top_dir)


class Project:
    def __init__(self, project_config, project_dir):
        self._config = project_config
        self._project_dir = project_dir

    def get_project_dir(self):
        return self._project_dir

    def add_custom_args_to_run_cli(self, cli_args_parser):
        if isinstance(self._config, HasCustomCliArgs):
            cli_group = cli_args_parser.add_argument_group("Project custom arguments")
            self._config.add_custom_cli_args(cli_group)

    def create_report_dir(self):
        return self._config.create_report_dir(self._project_dir)

    def get_suites(self, check_metadata_policy=True):
        suites = self._config.get_suites()
        if check_metadata_policy and isinstance(self._config, HasMetadataPolicy):
            policy = self._config.get_metadata_policy()
            policy.check_suites_compliance(suites)
        return suites

    def get_fixtures(self):
        return self._config.get_fixtures()

    def get_all_reporting_backends(self):
        return filter_available_reporting_backends(self._config.get_all_reporting_backends())

    def get_default_reporting_backends_for_test_run(self):
        return filter_available_reporting_backends(self._config.get_default_reporting_backends_for_test_run())

    def run_pre_session_hook(self, report_dir):
        if isinstance(self._config, HasPreRunHook):
            self._config.pre_run(report_dir)

    def run_post_session_hook(self, report_dir):
        if isinstance(self._config, HasPostRunHook):
            self._config.post_run(report_dir)

    def on_test_session_start(self, event):
        title = self._config.get_report_title()
        if title is not None:
            event.report.title = self._config.get_report_title()

        for key, value in self._config.get_report_info():
            event.report.add_info(key, value)


def _find_file_in_parent_directories(filename, dirname):
    if osp.exists(osp.join(dirname, filename)):
        return osp.join(dirname, filename)

    parent_dirname = osp.dirname(dirname)
    if parent_dirname == dirname:
        return None # root directory has been reached

    return _find_file_in_parent_directories(filename, parent_dirname)


def find_project_file():
    filename = os.environ.get("LCC_PROJECT_FILE")
    if filename is not None:
        return filename if osp.exists(filename) else None

    filename = _find_file_in_parent_directories(PROJECT_CONFIG_FILE, os.getcwd())
    return filename  # filename can be None


def find_project_dir():
    project_filename = find_project_file()
    if project_filename is None:
        return None
    return osp.dirname(project_filename)


def create_project(project_dir):
    shutil.copyfile(get_resource_path(osp.join("project", "template.py")), osp.join(project_dir, PROJECT_CONFIG_FILE))
    os.mkdir(osp.join(project_dir, "suites"))
    os.mkdir(osp.join(project_dir, "fixtures"))


def load_project_from_file(project_filename):
    project_dir = osp.dirname(project_filename)

    # load project module
    try:
        project_config_module = import_module(project_filename)
    except UserError as e:
        raise e # propagate UserError
    except Exception:
        raise ProjectError("Got an unexpected error while loading project:%s" % (
            serialize_current_exception()
        ))
    
    # get project config instance
    try:
        project_config = project_config_module.project
    except AttributeError:
        raise ProjectError("Cannot find symbol 'project' in module '%s'" % project_filename)
    if not isinstance(project_config, ProjectConfiguration):
        raise ProjectError("Symbol 'project' in module '%s' does not inherit lemoncheesecake.project.ProjectConfiguration" % project_filename)
    
    return Project(project_config, project_dir)


def load_project_from_dir(project_dir):
    return load_project_from_file(osp.join(project_dir, PROJECT_CONFIG_FILE))


def load_project():
    project_filename = find_project_file()
    if project_filename is None:
        raise ProjectError("Cannot find project file")

    return load_project_from_file(project_filename)



'''
Created on Sep 8, 2016

@author: nicolas
'''

import fnmatch
from functools import reduce

from lemoncheesecake.reporting import load_report
from lemoncheesecake.reporting.reportdir import DEFAULT_REPORT_DIR_NAME
from lemoncheesecake.testtree import flatten_tests
from lemoncheesecake.exceptions import UserError

__all__ = ("RunFilter", "ReportFilter", "filter_suites", "add_run_filter_cli_args",
           "make_run_filter")

NEGATIVE_FILTER_CHARS = "-^~"


def match_values(values, patterns):
    if not patterns:
        return True

    values = [value or "" for value in values]  # convert None to ""

    for pattern in patterns:
        if pattern[0] in NEGATIVE_FILTER_CHARS:
            if not fnmatch.filter(values, pattern[1:]):
                return True
        else:
            if fnmatch.filter(values, pattern):
                return True
    return False


def match_keyvalues(keyvalues, patterns):
    if not patterns:
        return True

    for key, value in patterns:
        if key in keyvalues:
            if value[0] in NEGATIVE_FILTER_CHARS:
                if not fnmatch.fnmatch(keyvalues[key], value[1:]):
                    return True
            else:
                if fnmatch.fnmatch(keyvalues[key], value):
                    return True
    return False


def match_values_lists(lsts, patterns):
    return match_values(
        reduce(lambda x, y: list(x) + list(y), lsts, []), # make a flat list
        patterns
    )


class Filter(object):
    def is_empty(self):
        raise NotImplemented()

    def is_test_disabled(self, test):
        raise NotImplemented()

    def match_test(self, test, suite):
        raise NotImplemented()


class BaseFilter(Filter):
    def __init__(self, paths=(), descriptions=(), tags=(), properties=(), links=(), enabled=False, disabled=False):
        self.paths = list(paths)
        self.descriptions = list(descriptions)
        self.tags = list(tags)
        self.properties = list(properties)
        self.links = list(links)
        self.enabled = enabled
        self.disabled = disabled

    def is_empty(self):
        return not any([
            self.paths, self.descriptions, self.tags, self.properties, self.links, self.enabled, self.disabled
        ])

    def is_test_disabled(self, test):
        return test.is_disabled()

    def match_test(self, test, suite):
        funcs = [
            lambda: self.is_test_disabled(test) if self.disabled else True,
            lambda: not self.is_test_disabled(test) if self.enabled else True,
            lambda: match_values(test.hierarchy_paths, self.paths),
            lambda: all(match_values(test.hierarchy_descriptions, descs) for descs in self.descriptions),
            lambda: all(match_values(test.hierarchy_tags, tags) for tags in self.tags),
            lambda: all(match_keyvalues(test.hierarchy_properties, props) for props in self.properties),
            lambda: all(match_values_lists(test.hierarchy_links, links) for links in self.links)
        ]
        return all(func() for func in funcs)


class RunFilter(BaseFilter):
    pass


class ReportFilter(RunFilter):
    def __init__(self, statuses=(), **kwargs):
        RunFilter.__init__(self, **kwargs)
        self.statuses = list(statuses)

    def is_empty(self):
        if not RunFilter.is_empty(self):
            return False

        return len(self.statuses) == 0

    def is_test_disabled(self, test):
        return test.status == "disabled"

    def match_test(self, test, suite):
        if not RunFilter.match_test(self, test, suite):
            return False

        if len(self.statuses) == 0:
            return True

        return test.status in self.statuses


class FromTestsFilter(Filter):
    def __init__(self, tests):
        self.tests = [test.path for test in tests]

    def is_empty(self):
        return False

    def is_test_disabled(self, test):
        return test.status == "disabled"

    def match_test(self, test, suite):
        return test.path in self.tests


def filter_suite(suite, filtr):
    filtered_suite = suite.pull_node()

    for test in suite.get_tests():
        if filtr.match_test(test, suite):
            filtered_suite.add_test(test.pull_node())

    for filtered_sub_suite in filter_suites(suite.get_suites(), filtr):
        filtered_suite.add_suite(filtered_sub_suite)

    return filtered_suite


def filter_suites(suites, filtr):
    filtered_suites = []

    for suite in suites:
        filtered_suite = filter_suite(suite, filtr)
        if not filtered_suite.is_empty():
            filtered_suites.append(filtered_suite)

    return filtered_suites


def add_run_filter_cli_args(cli_parser, no_positional_argument=False):
    def property_value(value):
        splitted = value.split(":")
        if len(splitted) != 2:
            raise ValueError()
        return splitted

    group = cli_parser.add_argument_group("Filtering")
    if no_positional_argument:
        group.add_argument(
            "--path", "-p", nargs="+", help="Filter on test/suite path (wildcard character '*' can be used)"
        )
    else:
        group.add_argument(
            "path", nargs="*", default=[], help="Filter on test/suite path (wildcard character '*' can be used)"
        )
    group.add_argument(
        "--desc", nargs="+", action="append", default=[], help="Filter on descriptions"
    )
    group.add_argument(
        "--tag", "-a", nargs="+", action="append", default=[], help="Filter on tags"
    )
    group.add_argument(
        "--property", "-m", nargs="+", type=property_value, action="append", default=[], help="Filter on properties"
    )
    group.add_argument(
        "--link", "-l", nargs="+", action="append", default=[], help="Filter on links (names and URLs)"
    )
    group.add_argument(
        "--disabled", action="store_true", help="Filter on disabled tests"
    )
    group.add_argument(
        "--passed", action="store_true", help="Filter on passed tests (implies/triggers --from-report)"
    )
    group.add_argument(
        "--failed", action="store_true", help="Filter on failed tests (implies/triggers --from-report)"
    )
    group.add_argument(
        "--skipped", action="store_true", help="Filter on skipped tests (implies/triggers --from-report)"
    )
    group.add_argument(
        "--enabled", action="store_true", help="Filter on enabled (non-disabled) tests"
    )
    group.add_argument(
        "--from-report", required=False, help="When enabled, the filtering is based on the given report"
    )

    return group

add_report_filter_cli_args = add_run_filter_cli_args


def _set_base_filter(fltr, cli_args):
    if cli_args.disabled and cli_args.enabled:
        raise UserError("--disabled and --enabled arguments are mutually exclusive")

    fltr.paths = cli_args.path
    fltr.descriptions = cli_args.desc
    fltr.tags = cli_args.tag
    fltr.properties = cli_args.property
    fltr.links = cli_args.link
    fltr.disabled = cli_args.disabled
    fltr.enabled = cli_args.enabled


def _set_run_filter(filtr, cli_args):
    if cli_args.passed or cli_args.failed or cli_args.skipped:
        raise UserError("--passed, --failed and --skipped arguments can only be used on the report-based filter")
    _set_base_filter(filtr, cli_args)


def _make_run_filter(cli_args):
    fltr = RunFilter()
    _set_run_filter(fltr, cli_args)
    return fltr


def _make_report_filter(cli_args):
    fltr = ReportFilter()
    _set_base_filter(fltr, cli_args)

    if cli_args.passed:
        fltr.statuses.append("passed")
    if cli_args.failed:
        fltr.statuses.append("failed")
    if cli_args.skipped:
        fltr.statuses.append("skipped")

    return fltr


def _make_from_report_filter(cli_args):
    report = load_report(cli_args.from_report or DEFAULT_REPORT_DIR_NAME)
    filtr = _make_report_filter(cli_args)
    suites = filter_suites(report.suites, filtr)
    return FromTestsFilter(flatten_tests(suites))


def make_run_filter(cli_args):
    if any([cli_args.from_report, cli_args.passed, cli_args.failed, cli_args.skipped]):
        return _make_from_report_filter(cli_args)
    else:
        return _make_run_filter(cli_args)


def make_report_filter(cli_args):
    if cli_args.from_report:
        return _make_from_report_filter(cli_args)
    else:
        return _make_report_filter(cli_args)
